import re
from threading import Timer as ThreadingTimer
from time import sleep
from unittest.mock import Mock, patch

import pytest
from freezegun import freeze_time

from moto.swf.exceptions import SWFDefaultUndefinedFault
from moto.swf.models import (
    ActivityType,
    Timeout,
    Timer,
    WorkflowExecution,
    WorkflowType,
)

from ..utils import (
    auto_start_decision_tasks,
    get_basic_domain,
    get_basic_workflow_type,
    make_workflow_execution,
)

VALID_ACTIVITY_TASK_ATTRIBUTES = {
    "activityId": "my-activity-001",
    "activityType": {"name": "test-activity", "version": "v1.1"},
    "taskList": {"name": "task-list-name"},
    "scheduleToStartTimeout": "600",
    "scheduleToCloseTimeout": "600",
    "startToCloseTimeout": "600",
    "heartbeatTimeout": "300",
}


def test_workflow_execution_creation():
    domain = get_basic_domain()
    wft = get_basic_workflow_type()
    wfe = WorkflowExecution(domain, wft, "ab1234", child_policy="TERMINATE")

    assert wfe.domain == domain
    assert wfe.workflow_type == wft
    assert wfe.child_policy == "TERMINATE"


def test_workflow_execution_creation_child_policy_logic():
    domain = get_basic_domain()

    assert (
        WorkflowExecution(
            domain,
            WorkflowType(
                "test-workflow",
                "v1.0",
                task_list="queue",
                default_child_policy="ABANDON",
                default_execution_start_to_close_timeout="300",
                default_task_start_to_close_timeout="300",
            ),
            "ab1234",
        ).child_policy
        == "ABANDON"
    )

    assert (
        WorkflowExecution(
            domain,
            WorkflowType(
                "test-workflow",
                "v1.0",
                task_list="queue",
                default_execution_start_to_close_timeout="300",
                default_task_start_to_close_timeout="300",
            ),
            "ab1234",
            child_policy="REQUEST_CANCEL",
        ).child_policy
        == "REQUEST_CANCEL"
    )

    with pytest.raises(SWFDefaultUndefinedFault):
        WorkflowExecution(domain, WorkflowType("test-workflow", "v1.0"), "ab1234")


def test_workflow_execution_string_representation():
    wfe = make_workflow_execution(child_policy="TERMINATE")
    assert re.match(r"^WorkflowExecution\(run_id: .*\)", str(wfe))


def test_workflow_execution_generates_a_random_run_id():
    domain = get_basic_domain()
    wft = get_basic_workflow_type()
    wfe1 = WorkflowExecution(domain, wft, "ab1234", child_policy="TERMINATE")
    wfe2 = WorkflowExecution(domain, wft, "ab1235", child_policy="TERMINATE")
    assert wfe1.run_id != wfe2.run_id


def test_workflow_execution_short_dict_representation():
    domain = get_basic_domain()
    wf_type = WorkflowType(
        "test-workflow",
        "v1.0",
        task_list="queue",
        default_child_policy="ABANDON",
        default_execution_start_to_close_timeout="300",
        default_task_start_to_close_timeout="300",
    )
    wfe = WorkflowExecution(domain, wf_type, "ab1234")

    sd = wfe.to_short_dict()
    assert sd["workflowId"] == "ab1234"
    assert "runId" in sd


def test_workflow_execution_medium_dict_representation():
    domain = get_basic_domain()
    wf_type = WorkflowType(
        "test-workflow",
        "v1.0",
        task_list="queue",
        default_child_policy="ABANDON",
        default_execution_start_to_close_timeout="300",
        default_task_start_to_close_timeout="300",
    )
    wfe = WorkflowExecution(domain, wf_type, "ab1234")

    md = wfe.to_medium_dict()
    assert md["execution"] == wfe.to_short_dict()
    assert md["workflowType"] == wf_type.to_short_dict()
    assert isinstance(md["startTimestamp"], float)
    assert md["executionStatus"] == "OPEN"
    assert md["cancelRequested"] is False
    assert "tagList" not in md

    wfe.tag_list = ["foo", "bar", "baz"]
    md = wfe.to_medium_dict()
    assert md["tagList"] == ["foo", "bar", "baz"]


def test_workflow_execution_full_dict_representation():
    domain = get_basic_domain()
    wf_type = WorkflowType(
        "test-workflow",
        "v1.0",
        task_list="queue",
        default_child_policy="ABANDON",
        default_execution_start_to_close_timeout="300",
        default_task_start_to_close_timeout="300",
    )
    wfe = WorkflowExecution(domain, wf_type, "ab1234")

    fd = wfe.to_full_dict()
    assert fd["executionInfo"] == wfe.to_medium_dict()
    assert fd["openCounts"]["openTimers"] == 0
    assert fd["openCounts"]["openDecisionTasks"] == 0
    assert fd["openCounts"]["openActivityTasks"] == 0
    assert fd["executionConfiguration"] == {
        "childPolicy": "ABANDON",
        "executionStartToCloseTimeout": "300",
        "taskList": {"name": "queue"},
        "taskStartToCloseTimeout": "300",
    }


def test_closed_workflow_execution_full_dict_representation():
    domain = get_basic_domain()
    wf_type = WorkflowType(
        "test-workflow",
        "v1.0",
        task_list="queue",
        default_child_policy="ABANDON",
        default_execution_start_to_close_timeout="300",
        default_task_start_to_close_timeout="300",
    )
    wfe = WorkflowExecution(domain, wf_type, "ab1234")
    wfe.execution_status = "CLOSED"
    wfe.close_status = "CANCELED"
    wfe.close_timestamp = 1420066801.123

    fd = wfe.to_full_dict()
    medium_dict = wfe.to_medium_dict()
    medium_dict["closeStatus"] = "CANCELED"
    medium_dict["closeTimestamp"] = 1420066801.123
    assert fd["executionInfo"] == medium_dict
    assert fd["openCounts"]["openTimers"] == 0
    assert fd["openCounts"]["openDecisionTasks"] == 0
    assert fd["openCounts"]["openActivityTasks"] == 0
    assert fd["executionConfiguration"] == {
        "childPolicy": "ABANDON",
        "executionStartToCloseTimeout": "300",
        "taskList": {"name": "queue"},
        "taskStartToCloseTimeout": "300",
    }


def test_workflow_execution_list_dict_representation():
    domain = get_basic_domain()
    wf_type = WorkflowType(
        "test-workflow",
        "v1.0",
        task_list="queue",
        default_child_policy="ABANDON",
        default_execution_start_to_close_timeout="300",
        default_task_start_to_close_timeout="300",
    )
    wfe = WorkflowExecution(domain, wf_type, "ab1234")

    ld = wfe.to_list_dict()
    assert ld["workflowType"]["version"] == "v1.0"
    assert ld["workflowType"]["name"] == "test-workflow"
    assert ld["executionStatus"] == "OPEN"
    assert ld["execution"]["workflowId"] == "ab1234"
    assert "runId" in ld["execution"]
    assert ld["cancelRequested"] is False
    assert "startTimestamp" in ld


def test_workflow_execution_schedule_decision_task():
    wfe = make_workflow_execution()
    assert wfe.open_counts["openDecisionTasks"] == 0
    wfe.schedule_decision_task()
    assert wfe.open_counts["openDecisionTasks"] == 1


def test_workflow_execution_dont_schedule_decision_if_existing_started_and_other_scheduled():
    wfe = make_workflow_execution()
    assert wfe.open_counts["openDecisionTasks"] == 0

    wfe.schedule_decision_task()
    assert wfe.open_counts["openDecisionTasks"] == 1

    wfe.decision_tasks[0].start("evt_id")

    wfe.schedule_decision_task()
    wfe.schedule_decision_task()
    assert wfe.open_counts["openDecisionTasks"] == 2


def test_workflow_execution_schedule_decision_if_existing_started_and_no_other_scheduled():
    wfe = make_workflow_execution()
    assert wfe.open_counts["openDecisionTasks"] == 0

    wfe.schedule_decision_task()
    assert wfe.open_counts["openDecisionTasks"] == 1

    wfe.decision_tasks[0].start("evt_id")

    wfe.schedule_decision_task()
    assert wfe.open_counts["openDecisionTasks"] == 2


def test_workflow_execution_start_decision_task():
    wfe = make_workflow_execution()
    wfe.schedule_decision_task()
    dt = wfe.decision_tasks[0]
    wfe.start_decision_task(dt.task_token, identity="srv01")
    dt = wfe.decision_tasks[0]
    assert dt.state == "STARTED"
    assert wfe.events()[-1].event_type == "DecisionTaskStarted"
    assert wfe.events()[-1].event_attributes["identity"] == "srv01"


def test_workflow_execution_history_events_ids():
    wfe = make_workflow_execution()
    wfe._add_event("WorkflowExecutionStarted")
    wfe._add_event("DecisionTaskScheduled")
    wfe._add_event("DecisionTaskStarted")
    ids = [evt.event_id for evt in wfe.events()]
    assert ids == [1, 2, 3]


@freeze_time("2015-01-01 12:00:00")
def test_workflow_execution_start():
    wfe = make_workflow_execution()
    assert wfe.events() == []

    wfe.start()
    assert wfe.start_timestamp == 1420113600.0
    assert len(wfe.events()) == 2
    assert wfe.events()[0].event_type == "WorkflowExecutionStarted"
    assert wfe.events()[1].event_type == "DecisionTaskScheduled"


@freeze_time("2015-01-02 12:00:00")
def test_workflow_execution_complete():
    wfe = make_workflow_execution()
    wfe.complete(123, result="foo")

    assert wfe.execution_status == "CLOSED"
    assert wfe.close_status == "COMPLETED"
    assert wfe.close_timestamp == 1420200000.0
    assert wfe.events()[-1].event_type == "WorkflowExecutionCompleted"
    assert wfe.events()[-1].event_attributes["decisionTaskCompletedEventId"] == 123
    assert wfe.events()[-1].event_attributes["result"] == "foo"


@freeze_time("2015-01-02 12:00:00")
def test_workflow_execution_fail():
    wfe = make_workflow_execution()
    wfe.fail(123, details="some details", reason="my rules")

    assert wfe.execution_status == "CLOSED"
    assert wfe.close_status == "FAILED"
    assert wfe.close_timestamp == 1420200000.0
    assert wfe.events()[-1].event_type == "WorkflowExecutionFailed"
    assert wfe.events()[-1].event_attributes["decisionTaskCompletedEventId"] == 123
    assert wfe.events()[-1].event_attributes["details"] == "some details"
    assert wfe.events()[-1].event_attributes["reason"] == "my rules"


@freeze_time("2015-01-01 12:00:00")
def test_workflow_execution_schedule_activity_task():
    wfe = make_workflow_execution()
    assert wfe.latest_activity_task_timestamp is None

    wfe.schedule_activity_task(123, VALID_ACTIVITY_TASK_ATTRIBUTES)

    assert wfe.latest_activity_task_timestamp == 1420113600.0

    assert wfe.open_counts["openActivityTasks"] == 1
    last_event = wfe.events()[-1]
    assert last_event.event_type == "ActivityTaskScheduled"
    assert last_event.event_attributes["decisionTaskCompletedEventId"] == 123
    assert last_event.event_attributes["taskList"]["name"] == "task-list-name"

    assert len(wfe.activity_tasks) == 1
    task = wfe.activity_tasks[0]
    assert task.activity_id == "my-activity-001"
    assert task.activity_type.name == "test-activity"
    assert task in wfe.domain.activity_task_lists["task-list-name"]


def test_workflow_execution_schedule_activity_task_without_task_list_should_take_default():
    wfe = make_workflow_execution()
    wfe.domain.add_type(ActivityType("test-activity", "v1.2", task_list="foobar"))
    wfe.schedule_activity_task(
        123,
        {
            "activityId": "my-activity-001",
            "activityType": {"name": "test-activity", "version": "v1.2"},
            "scheduleToStartTimeout": "600",
            "scheduleToCloseTimeout": "600",
            "startToCloseTimeout": "600",
            "heartbeatTimeout": "300",
        },
    )

    assert wfe.open_counts["openActivityTasks"] == 1
    last_event = wfe.events()[-1]
    assert last_event.event_type == "ActivityTaskScheduled"
    assert last_event.event_attributes["taskList"]["name"] == "foobar"

    task = wfe.activity_tasks[0]
    assert task in wfe.domain.activity_task_lists["foobar"]


def test_workflow_execution_schedule_activity_task_should_fail_if_wrong_attributes():
    wfe = make_workflow_execution()
    at = ActivityType("test-activity", "v1.1")
    at.status = "DEPRECATED"
    wfe.domain.add_type(at)
    wfe.domain.add_type(ActivityType("test-activity", "v1.2"))

    hsh = {
        "activityId": "my-activity-001",
        "activityType": {"name": "test-activity-does-not-exists", "version": "v1.1"},
    }

    wfe.schedule_activity_task(123, hsh)
    last_event = wfe.events()[-1]
    assert last_event.event_type == "ScheduleActivityTaskFailed"
    assert last_event.event_attributes["cause"] == "ACTIVITY_TYPE_DOES_NOT_EXIST"

    hsh["activityType"]["name"] = "test-activity"
    wfe.schedule_activity_task(123, hsh)
    last_event = wfe.events()[-1]
    assert last_event.event_type == "ScheduleActivityTaskFailed"
    assert last_event.event_attributes["cause"] == "ACTIVITY_TYPE_DEPRECATED"

    hsh["activityType"]["version"] = "v1.2"
    wfe.schedule_activity_task(123, hsh)
    last_event = wfe.events()[-1]
    assert last_event.event_type == "ScheduleActivityTaskFailed"
    assert last_event.event_attributes["cause"] == "DEFAULT_TASK_LIST_UNDEFINED"

    hsh["taskList"] = {"name": "foobar"}
    wfe.schedule_activity_task(123, hsh)
    last_event = wfe.events()[-1]
    assert last_event.event_type == "ScheduleActivityTaskFailed"
    assert last_event.event_attributes["cause"] == (
        "DEFAULT_SCHEDULE_TO_START_TIMEOUT_UNDEFINED"
    )

    hsh["scheduleToStartTimeout"] = "600"
    wfe.schedule_activity_task(123, hsh)
    last_event = wfe.events()[-1]
    assert last_event.event_type == "ScheduleActivityTaskFailed"
    assert last_event.event_attributes["cause"] == (
        "DEFAULT_SCHEDULE_TO_CLOSE_TIMEOUT_UNDEFINED"
    )

    hsh["scheduleToCloseTimeout"] = "600"
    wfe.schedule_activity_task(123, hsh)
    last_event = wfe.events()[-1]
    assert last_event.event_type == "ScheduleActivityTaskFailed"
    assert last_event.event_attributes["cause"] == (
        "DEFAULT_START_TO_CLOSE_TIMEOUT_UNDEFINED"
    )

    hsh["startToCloseTimeout"] = "600"
    wfe.schedule_activity_task(123, hsh)
    last_event = wfe.events()[-1]
    assert last_event.event_type == "ScheduleActivityTaskFailed"
    assert last_event.event_attributes["cause"] == (
        "DEFAULT_HEARTBEAT_TIMEOUT_UNDEFINED"
    )

    assert wfe.open_counts["openActivityTasks"] == 0
    assert len(wfe.activity_tasks) == 0
    assert len(wfe.domain.activity_task_lists) == 0

    hsh["heartbeatTimeout"] = "300"
    wfe.schedule_activity_task(123, hsh)
    last_event = wfe.events()[-1]
    assert last_event.event_type == "ActivityTaskScheduled"

    task = wfe.activity_tasks[0]
    assert task in wfe.domain.activity_task_lists["foobar"]
    assert wfe.open_counts["openDecisionTasks"] == 0
    assert wfe.open_counts["openActivityTasks"] == 1


def test_workflow_execution_schedule_activity_task_failure_triggers_new_decision():
    wfe = make_workflow_execution()
    wfe.start()
    task_token = wfe.decision_tasks[-1].task_token
    wfe.start_decision_task(task_token)
    wfe.complete_decision_task(
        task_token,
        execution_context="free-form execution context",
        decisions=[
            {
                "decisionType": "ScheduleActivityTask",
                "scheduleActivityTaskDecisionAttributes": {
                    "activityId": "my-activity-001",
                    "activityType": {
                        "name": "test-activity-does-not-exist",
                        "version": "v1.2",
                    },
                },
            },
            {
                "decisionType": "ScheduleActivityTask",
                "scheduleActivityTaskDecisionAttributes": {
                    "activityId": "my-activity-001",
                    "activityType": {
                        "name": "test-activity-does-not-exist",
                        "version": "v1.2",
                    },
                },
            },
        ],
    )

    assert wfe.latest_execution_context == "free-form execution context"
    assert wfe.open_counts["openActivityTasks"] == 0
    assert wfe.open_counts["openDecisionTasks"] == 1
    last_events = wfe.events()[-3:]
    assert last_events[0].event_type == "ScheduleActivityTaskFailed"
    assert last_events[1].event_type == "ScheduleActivityTaskFailed"
    assert last_events[2].event_type == "DecisionTaskScheduled"


def test_workflow_execution_schedule_activity_task_with_same_activity_id():
    wfe = make_workflow_execution()

    wfe.schedule_activity_task(123, VALID_ACTIVITY_TASK_ATTRIBUTES)
    assert wfe.open_counts["openActivityTasks"] == 1
    last_event = wfe.events()[-1]
    assert last_event.event_type == "ActivityTaskScheduled"

    wfe.schedule_activity_task(123, VALID_ACTIVITY_TASK_ATTRIBUTES)
    assert wfe.open_counts["openActivityTasks"] == 1
    last_event = wfe.events()[-1]
    assert last_event.event_type == "ScheduleActivityTaskFailed"
    assert last_event.event_attributes["cause"] == "ACTIVITY_ID_ALREADY_IN_USE"


def test_workflow_execution_start_activity_task():
    wfe = make_workflow_execution()
    wfe.schedule_activity_task(123, VALID_ACTIVITY_TASK_ATTRIBUTES)
    task_token = wfe.activity_tasks[-1].task_token
    wfe.start_activity_task(task_token, identity="worker01")
    task = wfe.activity_tasks[-1]
    assert task.state == "STARTED"
    assert wfe.events()[-1].event_type == "ActivityTaskStarted"
    assert wfe.events()[-1].event_attributes["identity"] == "worker01"


def test_complete_activity_task():
    wfe = make_workflow_execution()
    wfe.schedule_activity_task(123, VALID_ACTIVITY_TASK_ATTRIBUTES)
    task_token = wfe.activity_tasks[-1].task_token

    assert wfe.open_counts["openActivityTasks"] == 1
    assert wfe.open_counts["openDecisionTasks"] == 0

    wfe.start_activity_task(task_token, identity="worker01")
    wfe.complete_activity_task(task_token, result="a superb result")

    task = wfe.activity_tasks[-1]
    assert task.state == "COMPLETED"
    assert wfe.events()[-2].event_type == "ActivityTaskCompleted"
    assert wfe.events()[-1].event_type == "DecisionTaskScheduled"

    assert wfe.open_counts["openActivityTasks"] == 0
    assert wfe.open_counts["openDecisionTasks"] == 1


def test_terminate():
    wfe = make_workflow_execution()
    wfe.schedule_decision_task()
    wfe.terminate()

    assert wfe.execution_status == "CLOSED"
    assert wfe.close_status == "TERMINATED"
    assert wfe.close_cause == "OPERATOR_INITIATED"
    assert wfe.open_counts["openDecisionTasks"] == 1

    last_event = wfe.events()[-1]
    assert last_event.event_type == "WorkflowExecutionTerminated"
    # take default child_policy if not provided (as here)
    assert last_event.event_attributes["childPolicy"] == "ABANDON"


def test_first_timeout():
    wfe = make_workflow_execution()
    assert wfe.first_timeout() is None

    with freeze_time("2015-01-01 12:00:00"):
        wfe.start()
        assert wfe.first_timeout() is None

    with freeze_time("2015-01-01 14:01"):
        # 2 hours timeout reached
        assert isinstance(wfe.first_timeout(), Timeout)


# See moto/swf/models/workflow_execution.py "_process_timeouts()" for more
# details
def test_timeouts_are_processed_in_order_and_reevaluated():
    # Let's make a Workflow Execution with the following properties:
    # - execution start to close timeout of 8 mins
    # - (decision) task start to close timeout of 5 mins
    #
    # Now start the workflow execution, and look at the history 15 mins later:
    # - a first decision task is fired just after workflow execution start
    # - the first decision task should have timed out after 5 mins
    # - that fires a new decision task (which we hack to start automatically)
    # - then the workflow timeouts after 8 mins (shows gradual reevaluation)
    # - but the last scheduled decision task should *not* timeout (workflow closed)
    with freeze_time("2015-01-01 12:00:00"):
        wfe = make_workflow_execution(
            execution_start_to_close_timeout=8 * 60, task_start_to_close_timeout=5 * 60
        )
        # decision will automatically start
        wfe = auto_start_decision_tasks(wfe)
        wfe.start()
        event_idx = len(wfe.events())

    with freeze_time("2015-01-01 12:08:00"):
        wfe._process_timeouts()

        event_types = [e.event_type for e in wfe.events()[event_idx:]]
        assert event_types == [
            "DecisionTaskTimedOut",
            "DecisionTaskScheduled",
            "DecisionTaskStarted",
            "WorkflowExecutionTimedOut",
        ]


def test_record_marker():
    wfe = make_workflow_execution()
    MARKER_EVENT_ATTRIBUTES = {"markerName": "example_marker"}

    wfe.record_marker(123, MARKER_EVENT_ATTRIBUTES)

    last_event = wfe.events()[-1]
    assert last_event.event_type == "MarkerRecorded"
    assert last_event.event_attributes["markerName"] == "example_marker"
    assert last_event.event_attributes["decisionTaskCompletedEventId"] == 123


def test_start_timer():
    wfe = make_workflow_execution()
    START_TIMER_EVENT_ATTRIBUTES = {"startToFireTimeout": "10", "timerId": "abc123"}
    with patch("moto.swf.models.workflow_execution.ThreadingTimer"):
        wfe.start_timer(123, START_TIMER_EVENT_ATTRIBUTES)

        last_event = wfe.events()[-1]
        assert last_event.event_type == "TimerStarted"
        assert last_event.event_attributes["startToFireTimeout"] == "10"
        assert last_event.event_attributes["timerId"] == "abc123"
        assert last_event.event_attributes["decisionTaskCompletedEventId"] == 123


def test_start_timer_correctly_fires_timer_later():
    wfe = make_workflow_execution()
    START_TIMER_EVENT_ATTRIBUTES = {"startToFireTimeout": "60", "timerId": "abc123"}

    # Patch thread's event with one that immediately resolves
    with patch("threading.Event.wait"):
        wfe.start_timer(123, START_TIMER_EVENT_ATTRIBUTES)
        # Small wait to let both events populate
        sleep(0.5)

        second_to_last_event = wfe.events()[-2]
        last_event = wfe.events()[-1]
        assert second_to_last_event.event_type == "TimerFired"
        assert second_to_last_event.event_attributes["timerId"] == "abc123"
        assert second_to_last_event.event_attributes["startedEventId"] == 1
        assert last_event.event_type == "DecisionTaskScheduled"


def test_start_timer_fails_if_timer_already_started():
    wfe = make_workflow_execution()
    existing_timer = Mock(spec=ThreadingTimer)
    existing_timer.is_alive.return_value = True
    wfe._timers["abc123"] = Timer(existing_timer, 1)
    START_TIMER_EVENT_ATTRIBUTES = {"startToFireTimeout": "10", "timerId": "abc123"}

    wfe.start_timer(123, START_TIMER_EVENT_ATTRIBUTES)

    last_event = wfe.events()[-1]
    assert last_event.event_type == "StartTimerFailed"
    assert last_event.event_attributes["cause"] == "TIMER_ID_ALREADY_IN_USE"
    assert last_event.event_attributes["timerId"] == "abc123"
    assert last_event.event_attributes["decisionTaskCompletedEventId"] == 123


def test_cancel_timer():
    wfe = make_workflow_execution()
    existing_timer = Mock(spec=ThreadingTimer)
    existing_timer.is_alive.return_value = True
    wfe._timers["abc123"] = Timer(existing_timer, 1)

    wfe.cancel_timer(123, "abc123")

    last_event = wfe.events()[-1]
    assert last_event.event_type == "TimerCancelled"
    assert last_event.event_attributes["startedEventId"] == 1
    assert last_event.event_attributes["timerId"] == "abc123"
    assert last_event.event_attributes["decisionTaskCompletedEventId"] == 123
    existing_timer.cancel.assert_called_once()
    assert not wfe._timers.get("abc123")


def test_cancel_timer_fails_if_timer_not_found():
    wfe = make_workflow_execution()

    wfe.cancel_timer(123, "abc123")

    last_event = wfe.events()[-1]
    assert last_event.event_type == "CancelTimerFailed"
    assert last_event.event_attributes["cause"] == "TIMER_ID_UNKNOWN"
    assert last_event.event_attributes["decisionTaskCompletedEventId"] == 123


def test_cancel_workflow():
    wfe = make_workflow_execution()
    wfe.open_counts["openDecisionTasks"] = 1

    wfe.cancel(123, "I want to cancel")

    last_event = wfe.events()[-1]
    assert last_event.event_type == "WorkflowExecutionCanceled"
    assert last_event.event_attributes["details"] == "I want to cancel"
    assert last_event.event_attributes["decisionTaskCompletedEventId"] == 123


def test_cancel_workflow_fails_if_open_decision():
    wfe = make_workflow_execution()
    wfe.open_counts["openDecisionTasks"] = 2

    wfe.cancel(123, "I want to cancel")

    last_event = wfe.events()[-1]
    assert last_event.event_type == "CancelWorkflowExecutionFailed"
    assert last_event.event_attributes["cause"] == "UNHANDLED_DECISION"
    assert last_event.event_attributes["decisionTaskCompletedEventId"] == 123



from typing import Any, Dict, List, Type, Union

from moto.core.common_models import BaseModel

from .data_models import QuicksightGroup
from .exceptions import ParamValidationError, SchemaException, ValidationException

PAGINATION_MODEL = {
    "list_users": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,  # This should be the sum of the directory limits
        "unique_attribute": "arn",
    },
    "list_user_groups": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,  # This should be the sum of the directory limits
        "unique_attribute": "arn",
    },
    "list_groups": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,  # This should be the sum of the directory limits
        "unique_attribute": "arn",
    },
    "list_group_memberships": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,  # This should be the sum of the directory limits
        "unique_attribute": "arn",
    },
    "search_groups": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,  # This should be the sum of the directory limits
        "unique_attribute": "arn",
    },
}


class QuicksightBaseSearchFilter(BaseModel):
    """Base Search Filter."""

    schema: Dict[str, Any] = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
    }

    def __init__(self, operator: str, name: str, value: str):
        self.operator = operator
        self.name = name
        self.value = value

    def match(self, input: BaseModel) -> bool:
        raise NotImplementedError


class QuicksightGroupSearchFilter(QuicksightBaseSearchFilter):
    """Group Search Filter."""

    schema = {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "type": "array",
        "items": {
            "type": "object",
            "properties": {
                "Operator": {"type": "string", "enum": ["StringEquals", "StartsWith"]},
                "Name": {"type": "string", "enum": ["GROUP_NAME"]},
                "Value": {"type": "string"},
            },
            "required": ["Operator", "Name", "Value"],
        },
        "minItems": 1,
        "maxItems": 1,
    }

    def __init__(self, operator: str, name: str, value: str):
        super().__init__(operator, name, value)

    def match(self, input: BaseModel) -> bool:
        if isinstance(input, QuicksightGroup):
            if self.name == "GROUP_NAME" and self.operator == "StringEquals":
                return input.group_name.lower() == self.value.lower()
            if self.name == "GROUP_NAME" and self.operator == "StartsWith":
                return input.group_name.lower().startswith(self.value.lower())
        return False

    @classmethod
    def parse_filter(cls, filter: Dict[str, str]) -> QuicksightBaseSearchFilter:
        return QuicksightGroupSearchFilter(
            operator=filter.get("Operator", ""),
            name=filter.get("Name", ""),
            value=filter.get("Value", ""),
        )


class QuicksightSearchFilterList:
    """Generic QuickSight Search Filter List."""

    def __init__(self, filters: List[QuicksightBaseSearchFilter]):
        self.filters: List[QuicksightBaseSearchFilter] = filters

    def match(self, input: BaseModel) -> bool:
        return any([filter.match(input) for filter in self.filters])


class QuicksightSearchFilterFactory:
    """Creates the appropriate search filter."""

    @classmethod
    def validate_and_create_filter(
        cls, model_type: Type[BaseModel], input: Union[List[Dict[str, str]], None]
    ) -> QuicksightSearchFilterList:
        if issubclass(model_type, QuicksightGroup):
            if input is None:
                # Should never happen as there are other validations before but just in case.
                raise ParamValidationError(
                    'Missing required parameter in input: "Filters"'
                )
            from jsonschema import validate
            from jsonschema.exceptions import SchemaError, ValidationError

            try:
                validate(instance=input, schema=QuicksightGroupSearchFilter.schema)
                return QuicksightSearchFilterList(
                    filters=[
                        QuicksightGroupSearchFilter.parse_filter(filter)
                        for filter in input
                    ]
                )
            except ValidationError as err:
                raise ValidationException(err.message)

            except SchemaError as err:
                # This exception can happen only, when the schema defined in moto is wrong.
                # The schema is not presented by the user.
                raise SchemaException(f"Schema definition wrong in moto: {err.message}")

        raise NotImplementedError(
            f"Filter for '{model_type.__name__}' not implemented."
        )



"""Unit tests for s3tables-supported APIs."""

import boto3
import pytest

from moto import mock_aws


@mock_aws
def test_create_table_bucket():
    client = boto3.client("s3tables", region_name="us-east-1")
    response = client.create_table_bucket(name="foo")
    assert "arn" in response
    assert response["arn"].endswith("foo")


@mock_aws
def test_create_table_bucket_validates_name():
    client = boto3.client("s3tables", region_name="us-east-1")
    with pytest.raises(client.exceptions.BadRequestException) as exc:
        client.create_table_bucket(name="sthree-invalid-name")
    assert exc.match("bucket name is not valid")


@mock_aws
def test_list_table_buckets():
    client = boto3.client("s3tables", region_name="us-east-2")
    client.create_table_bucket(name="foo")
    response = client.list_table_buckets(prefix="foo")
    assert "tableBuckets" in response
    assert len(response["tableBuckets"]) == 1
    assert response["tableBuckets"][0]["name"] == "foo"


@mock_aws
def test_list_table_buckets_pagination():
    client = boto3.client("s3tables", region_name="us-east-2")
    client.create_table_bucket(name="foo")
    client.create_table_bucket(name="bar")
    response = client.list_table_buckets(maxBuckets=1)

    assert len(response["tableBuckets"]) == 1
    assert "continuationToken" in response

    response = client.list_table_buckets(
        maxBuckets=1, continuationToken=response["continuationToken"]
    )

    assert len(response["tableBuckets"]) == 1
    assert "continuationToken" not in response
    assert response["tableBuckets"][0]["name"] == "bar"


@mock_aws
def test_get_table_bucket():
    client = boto3.client("s3tables", region_name="us-east-1")
    response = client.create_table_bucket(name="foo")

    assert client.get_table_bucket(tableBucketARN=response["arn"])["name"] == "foo"


@mock_aws
def test_delete_table_bucket():
    client = boto3.client("s3tables", region_name="us-east-1")
    arn = client.create_table_bucket(name="foo")["arn"]
    response = client.list_table_buckets()
    assert response["tableBuckets"]

    client.delete_table_bucket(tableBucketARN=arn)
    response = client.list_table_buckets()
    assert not response["tableBuckets"]


@mock_aws
def test_create_namespace():
    client = boto3.client("s3tables", region_name="us-east-1")
    arn = client.create_table_bucket(name="foo")["arn"]
    resp = client.create_namespace(tableBucketARN=arn, namespace=["bar"])

    assert resp["namespace"] == ["bar"]


@mock_aws
def test_list_namespaces():
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["foo"])
    resp = client.list_namespaces(tableBucketARN=arn)
    assert resp["namespaces"]
    assert resp["namespaces"][0]["namespace"] == ["foo"]


@mock_aws
def test_list_namespaces_pagination():
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["foo"])
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    response = client.list_namespaces(tableBucketARN=arn, maxNamespaces=1)

    assert len(response["namespaces"]) == 1
    assert "continuationToken" in response

    response = client.list_namespaces(
        tableBucketARN=arn,
        maxNamespaces=1,
        continuationToken=response["continuationToken"],
    )

    assert len(response["namespaces"]) == 1
    assert "continuationToken" not in response
    assert response["namespaces"][0]["namespace"] == ["bar"]


@mock_aws
def test_get_namespace():
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    resp = client.get_namespace(tableBucketARN=arn, namespace="bar")
    assert resp["namespace"] == ["bar"]


@mock_aws
def test_delete_namespace():
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    assert client.list_namespaces(tableBucketARN=arn)["namespaces"]
    client.delete_namespace(tableBucketARN=arn, namespace="bar")
    assert not client.list_namespaces(tableBucketARN=arn)["namespaces"]


@mock_aws
def test_create_table():
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    resp = client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )
    assert "tableARN" in resp
    assert "versionToken" in resp


@mock_aws
def test_get_table():
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    table_arn = client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )["tableARN"]

    assert (
        client.get_table(tableBucketARN=arn, namespace="bar", name="baz")["tableARN"]
        == table_arn
    )


@mock_aws
def test_list_tables():
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )

    resp = client.list_tables(tableBucketARN=arn, namespace="bar")
    assert resp["tables"]
    assert resp["tables"][0]["name"] == "baz"


@mock_aws
def test_list_tables_pagination():
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )
    client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz2", format="ICEBERG"
    )

    resp = client.list_tables(tableBucketARN=arn, maxTables=1)
    assert len(resp["tables"]) == 1
    assert "continuationToken" in resp

    resp = client.list_tables(
        tableBucketARN=arn, maxTables=1, continuationToken=resp["continuationToken"]
    )

    assert len(resp["tables"]) == 1
    assert "continuationToken" not in resp
    assert resp["tables"][0]["name"] == "baz2"


@mock_aws
def test_delete_table():
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )

    resp = client.list_tables(tableBucketARN=arn)
    assert len(resp["tables"]) == 1

    client.delete_table(tableBucketARN=arn, namespace="bar", name="baz")

    resp = client.list_tables(tableBucketARN=arn)
    assert len(resp["tables"]) == 0


@mock_aws
def test_delete_table_deletes_underlying_table_storage():
    client = boto3.client("s3tables", region_name="us-east-2")
    s3 = boto3.client("s3", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )
    warehouse = client.get_table(tableBucketARN=arn, namespace="bar", name="baz")[
        "warehouseLocation"
    ]

    bucket_name = warehouse.replace("s3://", "")
    s3.head_bucket(Bucket=bucket_name)

    client.delete_table(tableBucketARN=arn, namespace="bar", name="baz")
    with pytest.raises(s3.exceptions.ClientError) as exc:
        s3.head_bucket(Bucket=bucket_name)
    exc.match("Not Found")


@mock_aws
def test_delete_table_fails_on_version_token_mismatch():
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    resp = client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )
    token = resp["versionToken"]

    with pytest.raises(client.exceptions.ConflictException) as exc:
        client.delete_table(
            tableBucketARN=arn, namespace="bar", name="baz", versionToken=f"{token}-foo"
        )
    exc.match("Provided version token does not match the table version token.")


@mock_aws
def test_get_table_metadata_location():
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )
    resp = client.get_table_metadata_location(
        tableBucketARN=arn, namespace="bar", name="baz"
    )
    assert "metadataLocation" not in resp
    assert "warehouseLocation" in resp


@mock_aws
def test_update_table_metadata_location():
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    resp = client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )
    resp = client.get_table(tableBucketARN=arn, namespace="bar", name="baz")
    warehouse_location = resp["warehouseLocation"]
    client.update_table_metadata_location(
        tableBucketARN=arn,
        namespace="bar",
        name="baz",
        metadataLocation=f"{warehouse_location}/abc",
        versionToken=resp["versionToken"],
    )
    resp = client.get_table_metadata_location(
        tableBucketARN=arn, namespace="bar", name="baz"
    )
    assert "metadataLocation" in resp
    assert resp["metadataLocation"] == f"{warehouse_location}/abc"


@mock_aws
def test_update_table_metadata_location_raises_exception_on_invalid_path():
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    resp = client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )
    with pytest.raises(client.exceptions.BadRequestException) as exc:
        client.update_table_metadata_location(
            tableBucketARN=arn,
            namespace="bar",
            name="baz",
            metadataLocation="s3://abc",
            versionToken=resp["versionToken"],
        )
    exc.match("The specified metadata location is not valid.")


@mock_aws
def test_write_metadata_to_table() -> None:
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    resp = client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )

    resp = client.get_table(tableBucketARN=arn, namespace="bar", name="baz")
    s3 = boto3.client("s3", region_name="us-east-2")
    metadata = b'{"foo":"bar"}'

    bucket_name = resp["warehouseLocation"].replace("s3://", "")
    key = "metadata/00001-foo.metadata.json"
    s3.put_object(Bucket=bucket_name, Key=key, Body=metadata)

    client.update_table_metadata_location(
        tableBucketARN=arn,
        namespace="bar",
        name="baz",
        metadataLocation=f"s3://{bucket_name}/{key}",
        versionToken=resp["versionToken"],
    )
    resp = client.get_table_metadata_location(
        tableBucketARN=arn, namespace="bar", name="baz"
    )
    bucket, key = resp["metadataLocation"].replace("s3://", "").split("/", 1)
    resp = s3.get_object(Bucket=bucket, Key=key)
    assert resp["Body"].read() == metadata


@mock_aws
def test_underlying_table_storage_does_not_support_list_objects() -> None:
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    resp = client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )

    resp = client.get_table(tableBucketARN=arn, namespace="bar", name="baz")
    s3 = boto3.client("s3", region_name="us-east-2")

    bucket_name = resp["warehouseLocation"].replace("s3://", "")
    with pytest.raises(s3.exceptions.ClientError) as exc:
        s3.list_objects_v2(Bucket=bucket_name)
    assert exc.match("The specified method is not allowed against this resource.")


@mock_aws
def test_underlying_table_storage_does_not_support_delete_object() -> None:
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    resp = client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )

    resp = client.get_table(tableBucketARN=arn, namespace="bar", name="baz")
    s3 = boto3.client("s3", region_name="us-east-2")

    bucket_name = resp["warehouseLocation"].replace("s3://", "")
    s3.put_object(Bucket=bucket_name, Key="test", Body=b"{}")
    with pytest.raises(s3.exceptions.ClientError) as exc:
        s3.delete_object(Bucket=bucket_name, Key="test")
    assert exc.match("The specified method is not allowed against this resource.")


@mock_aws
def test_rename_table() -> None:
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    resp = client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )

    client.create_namespace(tableBucketARN=arn, namespace=["bar-two"])
    client.rename_table(
        tableBucketARN=arn,
        namespace="bar",
        name="baz",
        newNamespaceName="bar-two",
        newName="baz-two",
        versionToken=resp["versionToken"],
    )
    assert (
        client.get_table(tableBucketARN=arn, namespace="bar-two", name="baz-two")[
            "name"
        ]
        == "baz-two"
    )
    assert client.get_table(tableBucketARN=arn, namespace="bar-two", name="baz-two")[
        "namespace"
    ] == ["bar-two"]


@mock_aws
def test_rename_table_fails_when_destination_namespace_does_not_exist() -> None:
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    resp = client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )

    with pytest.raises(client.exceptions.NotFoundException) as ctx:
        client.rename_table(
            tableBucketARN=arn,
            namespace="bar",
            name="baz",
            newNamespaceName="bar-two",
            newName="baz-two",
            versionToken=resp["versionToken"],
        )
    assert ctx.match("The specified destination namespace does not exist.")


@mock_aws
def test_rename_table_fails_when_no_updates_are_specified() -> None:
    client = boto3.client("s3tables", region_name="us-east-2")
    arn = client.create_table_bucket(name="foo")["arn"]
    client.create_namespace(tableBucketARN=arn, namespace=["bar"])
    resp = client.create_table(
        tableBucketARN=arn, namespace="bar", name="baz", format="ICEBERG"
    )

    with pytest.raises(client.exceptions.BadRequestException) as ctx:
        client.rename_table(
            tableBucketARN=arn,
            namespace="bar",
            name="baz",
            versionToken=resp["versionToken"],
        )
    assert ctx.match("Neither a new namespace name nor a new table name is specified.")



from typing import Tuple

from antlr4 import CommonTokenStream, InputStream, ParserRuleContext

from moto.stepfunctions.parser.asl.antlr.runtime.ASLLexer import ASLLexer
from moto.stepfunctions.parser.asl.antlr.runtime.ASLParser import ASLParser
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.parse.asl_parser import (
    AmazonStateLanguageParser,
    ASLParserException,
    SyntaxErrorListener,
)
from moto.stepfunctions.parser.asl.parse.test_state.preprocessor import (
    TestStatePreprocessor,
)


class TestStateAmazonStateLanguageParser(AmazonStateLanguageParser):
    @staticmethod
    def parse(definition: str) -> Tuple[EvalComponent, ParserRuleContext]:
        # Attempt to build the AST and look out for syntax errors.
        syntax_error_listener = SyntaxErrorListener()

        input_stream = InputStream(definition)
        lexer = ASLLexer(input_stream)
        stream = CommonTokenStream(lexer)
        parser = ASLParser(stream)
        parser.removeErrorListeners()
        parser.addErrorListener(syntax_error_listener)
        # Unlike the main Program parser, TestState parsing occurs at a state declaration level.
        tree = parser.state_decl_body()

        errors = syntax_error_listener.errors
        if errors:
            raise ASLParserException(errors=errors)

        # Attempt to preprocess the AST into evaluation components.
        preprocessor = TestStatePreprocessor()
        test_state_program = preprocessor.visit(tree)

        return test_state_program, tree



import boto3
import pytest

from moto import mock_aws
from moto.utilities.id_generator import ResourceIdentifier, moto_id_manager


@pytest.fixture(scope="function")
def account_id():
    """Return the current Account ID. Will reach out to AWS if so configured."""
    from . import allow_aws_request

    if allow_aws_request():
        identity = boto3.client("sts", "us-east-1").get_caller_identity()
        yield identity["Account"]
    else:
        with mock_aws():
            identity = boto3.client("sts", "us-east-1").get_caller_identity()
            yield identity["Account"]


@pytest.fixture
def set_custom_id():
    set_ids = []

    def _set_custom_id(resource_identifier: ResourceIdentifier, custom_id):
        moto_id_manager.set_custom_id(
            resource_identifier=resource_identifier, custom_id=custom_id
        )
        set_ids.append(resource_identifier)

    yield _set_custom_id

    for resource_identifier in set_ids:
        moto_id_manager.unset_custom_id(resource_identifier)



import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

from .test_elbv2 import create_load_balancer


@mock_aws
def test_create_listener_with_tags():
    response, _, _, _, _, elbv2 = create_load_balancer()

    load_balancer_arn = response.get("LoadBalancers")[0].get("LoadBalancerArn")

    listener_arn = elbv2.create_listener(
        LoadBalancerArn=load_balancer_arn,
        Protocol="HTTP",
        Port=80,
        DefaultActions=[],
        Tags=[{"Key": "k1", "Value": "v1"}],
    )["Listeners"][0]["ListenerArn"]

    # Ensure the tags persisted
    response = elbv2.describe_tags(ResourceArns=[listener_arn])
    tags = {d["Key"]: d["Value"] for d in response["TagDescriptions"][0]["Tags"]}
    assert tags == {"k1": "v1"}


@mock_aws
def test_listener_add_remove_tags():
    response, _, _, _, _, elbv2 = create_load_balancer()

    load_balancer_arn = response.get("LoadBalancers")[0].get("LoadBalancerArn")

    listener_arn = elbv2.create_listener(
        LoadBalancerArn=load_balancer_arn,
        Protocol="HTTP",
        Port=80,
        DefaultActions=[],
        Tags=[{"Key": "k1", "Value": "v1"}],
    )["Listeners"][0]["ListenerArn"]

    elbv2.add_tags(ResourceArns=[listener_arn], Tags=[{"Key": "a", "Value": "b"}])

    tags = elbv2.describe_tags(ResourceArns=[listener_arn])["TagDescriptions"][0][
        "Tags"
    ]
    assert tags == [{"Key": "k1", "Value": "v1"}, {"Key": "a", "Value": "b"}]

    elbv2.add_tags(
        ResourceArns=[listener_arn],
        Tags=[
            {"Key": "a", "Value": "b"},
            {"Key": "b", "Value": "b"},
            {"Key": "c", "Value": "b"},
        ],
    )

    tags = elbv2.describe_tags(ResourceArns=[listener_arn])["TagDescriptions"][0][
        "Tags"
    ]
    assert tags == [
        {"Key": "k1", "Value": "v1"},
        {"Key": "a", "Value": "b"},
        {"Key": "b", "Value": "b"},
        {"Key": "c", "Value": "b"},
    ]

    elbv2.remove_tags(ResourceArns=[listener_arn], TagKeys=["a"])

    tags = elbv2.describe_tags(ResourceArns=[listener_arn])["TagDescriptions"][0][
        "Tags"
    ]
    assert tags == [
        {"Key": "k1", "Value": "v1"},
        {"Key": "b", "Value": "b"},
        {"Key": "c", "Value": "b"},
    ]


@mock_aws
def test_remove_tags_to_invalid_listener():
    response, _, _, _, _, elbv2 = create_load_balancer()
    load_balancer_arn = response.get("LoadBalancers")[0].get("LoadBalancerArn")

    listener_arn = elbv2.create_listener(
        LoadBalancerArn=load_balancer_arn,
        Protocol="HTTP",
        Port=80,
        DefaultActions=[],
    )["Listeners"][0]["ListenerArn"]

    # add a random string in the ARN to make the resource non-existent
    BAD_ARN = listener_arn + "randomstring"

    # test for exceptions on remove tags
    with pytest.raises(ClientError) as err:
        elbv2.remove_tags(ResourceArns=[BAD_ARN], TagKeys=["a"])

    assert err.value.response["Error"]["Code"] == "ListenerNotFound"
    assert (
        err.value.response["Error"]["Message"]
        == "The specified listener does not exist."
    )



"""workspacesweb base URL and path."""

from .responses import WorkSpacesWebResponse

url_bases = [
    r"https?://workspaces-web\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/browserSettings$": WorkSpacesWebResponse.dispatch,
    "{0}/networkSettings$": WorkSpacesWebResponse.dispatch,
    "{0}/userSettings$": WorkSpacesWebResponse.dispatch,
    "{0}/userAccessLoggingSettings$": WorkSpacesWebResponse.dispatch,
    "{0}/portals$": WorkSpacesWebResponse.dispatch,
    "{0}/portals/(?P<portalArn>[^/]+)portal/(?P<uuid>[^/]+)$": WorkSpacesWebResponse.portal,
    "{0}/browserSettings/(?P<browserSettingsArn>.+)$": WorkSpacesWebResponse.browser_settings,
    "{0}/networkSettings/(?P<networkSettingsArn>.+)$": WorkSpacesWebResponse.network_settings,
    "{0}/userSettings/(?P<userSettingsArn>.+)$": WorkSpacesWebResponse.user_settings,
    "{0}/userAccessLoggingSettings/(?P<userAccessLoggingSettingsArn>.+)$": WorkSpacesWebResponse.user_access_logging_settings,
    "{0}/portals/(?P<portalArn>.*)/browserSettings$": WorkSpacesWebResponse.dispatch,
    "{0}/portals/(?P<portalArn>.*)/networkSettings$": WorkSpacesWebResponse.dispatch,
    "{0}/portals/(?P<portalArn>.*)/userSettings$": WorkSpacesWebResponse.dispatch,
    "{0}/portals/(?P<portalArn>.*)/userAccessLoggingSettings$": WorkSpacesWebResponse.dispatch,
    "{0}/tags/(?P<resourceArn>.+)$": WorkSpacesWebResponse.dispatch,
}



import os
from functools import wraps
from uuid import uuid4

import boto3
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.s3.responses import DEFAULT_REGION_NAME


def s3_aws_verified(func):
    """
    Function that is verified to work against AWS.
    Can be run against AWS at any time by setting:
      MOTO_TEST_ALLOW_AWS_REQUEST=true

    If this environment variable is not set, the function runs in a `mock_aws` context.

    This decorator will:
      - Create a bucket
      - Run the test and pass the bucket_name as an argument
      - Delete the objects and the bucket itself
    """

    @wraps(func)
    def pagination_wrapper(**kwargs):
        bucket_name = str(uuid4())

        allow_aws_request = (
            os.environ.get("MOTO_TEST_ALLOW_AWS_REQUEST", "false").lower() == "true"
        )

        if allow_aws_request:
            print(f"Test {func} will create {bucket_name}")  # noqa: T201
            resp = create_bucket_and_test(bucket_name, **kwargs)
        else:
            with mock_aws():
                resp = create_bucket_and_test(bucket_name, **kwargs)
        return resp

    def create_bucket_and_test(bucket_name, **kwargs):
        client = boto3.client("s3", region_name=DEFAULT_REGION_NAME)

        client.create_bucket(Bucket=bucket_name)
        client.put_bucket_tagging(
            Bucket=bucket_name,
            Tagging={"TagSet": [{"Key": "environment", "Value": "moto_tests"}]},
        )
        try:
            resp = func(**kwargs, bucket_name=bucket_name)
        finally:
            ### CLEANUP ###

            empty_bucket(client, bucket_name)
            client.delete_bucket(Bucket=bucket_name)

        return resp

    return pagination_wrapper


def empty_bucket(client, bucket_name):
    # Delete any object lock config, if set before
    try:
        client.get_object_lock_configuration(Bucket=bucket_name)
        kwargs = {"BypassGovernanceRetention": True}
    except ClientError:
        # No ObjectLock set
        kwargs = {}

    versions = client.list_object_versions(Bucket=bucket_name).get("Versions", [])
    for key in versions:
        client.delete_object(
            Bucket=bucket_name, Key=key["Key"], VersionId=key.get("VersionId"), **kwargs
        )
    delete_markers = client.list_object_versions(Bucket=bucket_name).get(
        "DeleteMarkers", []
    )
    for key in delete_markers:
        client.delete_object(
            Bucket=bucket_name, Key=key["Key"], VersionId=key.get("VersionId"), **kwargs
        )


def generate_content_md5(content: bytes) -> str:
    import base64
    import hashlib

    md = hashlib.md5(content).digest()
    content_md5 = base64.b64encode(md).decode("utf-8")
    return content_md5



# Generated from ASLParser.g4 by ANTLR 4.13.2
from antlr4 import ParseTreeListener

if "." in __name__:
    from .ASLParser import ASLParser
else:
    from ASLParser import ASLParser


# This class defines a complete listener for a parse tree produced by ASLParser.
class ASLParserListener(ParseTreeListener):
    # Enter a parse tree produced by ASLParser#state_machine.
    def enterState_machine(self, ctx: ASLParser.State_machineContext):
        pass

    # Exit a parse tree produced by ASLParser#state_machine.
    def exitState_machine(self, ctx: ASLParser.State_machineContext):
        pass

    # Enter a parse tree produced by ASLParser#program_decl.
    def enterProgram_decl(self, ctx: ASLParser.Program_declContext):
        pass

    # Exit a parse tree produced by ASLParser#program_decl.
    def exitProgram_decl(self, ctx: ASLParser.Program_declContext):
        pass

    # Enter a parse tree produced by ASLParser#top_layer_stmt.
    def enterTop_layer_stmt(self, ctx: ASLParser.Top_layer_stmtContext):
        pass

    # Exit a parse tree produced by ASLParser#top_layer_stmt.
    def exitTop_layer_stmt(self, ctx: ASLParser.Top_layer_stmtContext):
        pass

    # Enter a parse tree produced by ASLParser#startat_decl.
    def enterStartat_decl(self, ctx: ASLParser.Startat_declContext):
        pass

    # Exit a parse tree produced by ASLParser#startat_decl.
    def exitStartat_decl(self, ctx: ASLParser.Startat_declContext):
        pass

    # Enter a parse tree produced by ASLParser#comment_decl.
    def enterComment_decl(self, ctx: ASLParser.Comment_declContext):
        pass

    # Exit a parse tree produced by ASLParser#comment_decl.
    def exitComment_decl(self, ctx: ASLParser.Comment_declContext):
        pass

    # Enter a parse tree produced by ASLParser#version_decl.
    def enterVersion_decl(self, ctx: ASLParser.Version_declContext):
        pass

    # Exit a parse tree produced by ASLParser#version_decl.
    def exitVersion_decl(self, ctx: ASLParser.Version_declContext):
        pass

    # Enter a parse tree produced by ASLParser#query_language_decl.
    def enterQuery_language_decl(self, ctx: ASLParser.Query_language_declContext):
        pass

    # Exit a parse tree produced by ASLParser#query_language_decl.
    def exitQuery_language_decl(self, ctx: ASLParser.Query_language_declContext):
        pass

    # Enter a parse tree produced by ASLParser#state_stmt.
    def enterState_stmt(self, ctx: ASLParser.State_stmtContext):
        pass

    # Exit a parse tree produced by ASLParser#state_stmt.
    def exitState_stmt(self, ctx: ASLParser.State_stmtContext):
        pass

    # Enter a parse tree produced by ASLParser#states_decl.
    def enterStates_decl(self, ctx: ASLParser.States_declContext):
        pass

    # Exit a parse tree produced by ASLParser#states_decl.
    def exitStates_decl(self, ctx: ASLParser.States_declContext):
        pass

    # Enter a parse tree produced by ASLParser#state_name.
    def enterState_name(self, ctx: ASLParser.State_nameContext):
        pass

    # Exit a parse tree produced by ASLParser#state_name.
    def exitState_name(self, ctx: ASLParser.State_nameContext):
        pass

    # Enter a parse tree produced by ASLParser#state_decl.
    def enterState_decl(self, ctx: ASLParser.State_declContext):
        pass

    # Exit a parse tree produced by ASLParser#state_decl.
    def exitState_decl(self, ctx: ASLParser.State_declContext):
        pass

    # Enter a parse tree produced by ASLParser#state_decl_body.
    def enterState_decl_body(self, ctx: ASLParser.State_decl_bodyContext):
        pass

    # Exit a parse tree produced by ASLParser#state_decl_body.
    def exitState_decl_body(self, ctx: ASLParser.State_decl_bodyContext):
        pass

    # Enter a parse tree produced by ASLParser#type_decl.
    def enterType_decl(self, ctx: ASLParser.Type_declContext):
        pass

    # Exit a parse tree produced by ASLParser#type_decl.
    def exitType_decl(self, ctx: ASLParser.Type_declContext):
        pass

    # Enter a parse tree produced by ASLParser#next_decl.
    def enterNext_decl(self, ctx: ASLParser.Next_declContext):
        pass

    # Exit a parse tree produced by ASLParser#next_decl.
    def exitNext_decl(self, ctx: ASLParser.Next_declContext):
        pass

    # Enter a parse tree produced by ASLParser#resource_decl.
    def enterResource_decl(self, ctx: ASLParser.Resource_declContext):
        pass

    # Exit a parse tree produced by ASLParser#resource_decl.
    def exitResource_decl(self, ctx: ASLParser.Resource_declContext):
        pass

    # Enter a parse tree produced by ASLParser#input_path_decl_var.
    def enterInput_path_decl_var(self, ctx: ASLParser.Input_path_decl_varContext):
        pass

    # Exit a parse tree produced by ASLParser#input_path_decl_var.
    def exitInput_path_decl_var(self, ctx: ASLParser.Input_path_decl_varContext):
        pass

    # Enter a parse tree produced by ASLParser#input_path_decl_path_context_object.
    def enterInput_path_decl_path_context_object(
        self, ctx: ASLParser.Input_path_decl_path_context_objectContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#input_path_decl_path_context_object.
    def exitInput_path_decl_path_context_object(
        self, ctx: ASLParser.Input_path_decl_path_context_objectContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#input_path_decl_path.
    def enterInput_path_decl_path(self, ctx: ASLParser.Input_path_decl_pathContext):
        pass

    # Exit a parse tree produced by ASLParser#input_path_decl_path.
    def exitInput_path_decl_path(self, ctx: ASLParser.Input_path_decl_pathContext):
        pass

    # Enter a parse tree produced by ASLParser#result_decl.
    def enterResult_decl(self, ctx: ASLParser.Result_declContext):
        pass

    # Exit a parse tree produced by ASLParser#result_decl.
    def exitResult_decl(self, ctx: ASLParser.Result_declContext):
        pass

    # Enter a parse tree produced by ASLParser#result_path_decl.
    def enterResult_path_decl(self, ctx: ASLParser.Result_path_declContext):
        pass

    # Exit a parse tree produced by ASLParser#result_path_decl.
    def exitResult_path_decl(self, ctx: ASLParser.Result_path_declContext):
        pass

    # Enter a parse tree produced by ASLParser#output_path_decl_var.
    def enterOutput_path_decl_var(self, ctx: ASLParser.Output_path_decl_varContext):
        pass

    # Exit a parse tree produced by ASLParser#output_path_decl_var.
    def exitOutput_path_decl_var(self, ctx: ASLParser.Output_path_decl_varContext):
        pass

    # Enter a parse tree produced by ASLParser#output_path_decl_path_context_object.
    def enterOutput_path_decl_path_context_object(
        self, ctx: ASLParser.Output_path_decl_path_context_objectContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#output_path_decl_path_context_object.
    def exitOutput_path_decl_path_context_object(
        self, ctx: ASLParser.Output_path_decl_path_context_objectContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#output_path_decl_path.
    def enterOutput_path_decl_path(self, ctx: ASLParser.Output_path_decl_pathContext):
        pass

    # Exit a parse tree produced by ASLParser#output_path_decl_path.
    def exitOutput_path_decl_path(self, ctx: ASLParser.Output_path_decl_pathContext):
        pass

    # Enter a parse tree produced by ASLParser#end_decl.
    def enterEnd_decl(self, ctx: ASLParser.End_declContext):
        pass

    # Exit a parse tree produced by ASLParser#end_decl.
    def exitEnd_decl(self, ctx: ASLParser.End_declContext):
        pass

    # Enter a parse tree produced by ASLParser#default_decl.
    def enterDefault_decl(self, ctx: ASLParser.Default_declContext):
        pass

    # Exit a parse tree produced by ASLParser#default_decl.
    def exitDefault_decl(self, ctx: ASLParser.Default_declContext):
        pass

    # Enter a parse tree produced by ASLParser#error_jsonata.
    def enterError_jsonata(self, ctx: ASLParser.Error_jsonataContext):
        pass

    # Exit a parse tree produced by ASLParser#error_jsonata.
    def exitError_jsonata(self, ctx: ASLParser.Error_jsonataContext):
        pass

    # Enter a parse tree produced by ASLParser#error_string.
    def enterError_string(self, ctx: ASLParser.Error_stringContext):
        pass

    # Exit a parse tree produced by ASLParser#error_string.
    def exitError_string(self, ctx: ASLParser.Error_stringContext):
        pass

    # Enter a parse tree produced by ASLParser#error_path_decl_var.
    def enterError_path_decl_var(self, ctx: ASLParser.Error_path_decl_varContext):
        pass

    # Exit a parse tree produced by ASLParser#error_path_decl_var.
    def exitError_path_decl_var(self, ctx: ASLParser.Error_path_decl_varContext):
        pass

    # Enter a parse tree produced by ASLParser#error_path_decl_path.
    def enterError_path_decl_path(self, ctx: ASLParser.Error_path_decl_pathContext):
        pass

    # Exit a parse tree produced by ASLParser#error_path_decl_path.
    def exitError_path_decl_path(self, ctx: ASLParser.Error_path_decl_pathContext):
        pass

    # Enter a parse tree produced by ASLParser#error_path_decl_context.
    def enterError_path_decl_context(
        self, ctx: ASLParser.Error_path_decl_contextContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#error_path_decl_context.
    def exitError_path_decl_context(
        self, ctx: ASLParser.Error_path_decl_contextContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#error_path_decl_intrinsic.
    def enterError_path_decl_intrinsic(
        self, ctx: ASLParser.Error_path_decl_intrinsicContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#error_path_decl_intrinsic.
    def exitError_path_decl_intrinsic(
        self, ctx: ASLParser.Error_path_decl_intrinsicContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#cause_jsonata.
    def enterCause_jsonata(self, ctx: ASLParser.Cause_jsonataContext):
        pass

    # Exit a parse tree produced by ASLParser#cause_jsonata.
    def exitCause_jsonata(self, ctx: ASLParser.Cause_jsonataContext):
        pass

    # Enter a parse tree produced by ASLParser#cause_string.
    def enterCause_string(self, ctx: ASLParser.Cause_stringContext):
        pass

    # Exit a parse tree produced by ASLParser#cause_string.
    def exitCause_string(self, ctx: ASLParser.Cause_stringContext):
        pass

    # Enter a parse tree produced by ASLParser#cause_path_decl_var.
    def enterCause_path_decl_var(self, ctx: ASLParser.Cause_path_decl_varContext):
        pass

    # Exit a parse tree produced by ASLParser#cause_path_decl_var.
    def exitCause_path_decl_var(self, ctx: ASLParser.Cause_path_decl_varContext):
        pass

    # Enter a parse tree produced by ASLParser#cause_path_decl_path.
    def enterCause_path_decl_path(self, ctx: ASLParser.Cause_path_decl_pathContext):
        pass

    # Exit a parse tree produced by ASLParser#cause_path_decl_path.
    def exitCause_path_decl_path(self, ctx: ASLParser.Cause_path_decl_pathContext):
        pass

    # Enter a parse tree produced by ASLParser#cause_path_decl_context.
    def enterCause_path_decl_context(
        self, ctx: ASLParser.Cause_path_decl_contextContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#cause_path_decl_context.
    def exitCause_path_decl_context(
        self, ctx: ASLParser.Cause_path_decl_contextContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#cause_path_decl_intrinsic.
    def enterCause_path_decl_intrinsic(
        self, ctx: ASLParser.Cause_path_decl_intrinsicContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#cause_path_decl_intrinsic.
    def exitCause_path_decl_intrinsic(
        self, ctx: ASLParser.Cause_path_decl_intrinsicContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#seconds_jsonata.
    def enterSeconds_jsonata(self, ctx: ASLParser.Seconds_jsonataContext):
        pass

    # Exit a parse tree produced by ASLParser#seconds_jsonata.
    def exitSeconds_jsonata(self, ctx: ASLParser.Seconds_jsonataContext):
        pass

    # Enter a parse tree produced by ASLParser#seconds_int.
    def enterSeconds_int(self, ctx: ASLParser.Seconds_intContext):
        pass

    # Exit a parse tree produced by ASLParser#seconds_int.
    def exitSeconds_int(self, ctx: ASLParser.Seconds_intContext):
        pass

    # Enter a parse tree produced by ASLParser#seconds_path_decl_var.
    def enterSeconds_path_decl_var(self, ctx: ASLParser.Seconds_path_decl_varContext):
        pass

    # Exit a parse tree produced by ASLParser#seconds_path_decl_var.
    def exitSeconds_path_decl_var(self, ctx: ASLParser.Seconds_path_decl_varContext):
        pass

    # Enter a parse tree produced by ASLParser#seconds_path_decl_value.
    def enterSeconds_path_decl_value(
        self, ctx: ASLParser.Seconds_path_decl_valueContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#seconds_path_decl_value.
    def exitSeconds_path_decl_value(
        self, ctx: ASLParser.Seconds_path_decl_valueContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#timestamp_jsonata.
    def enterTimestamp_jsonata(self, ctx: ASLParser.Timestamp_jsonataContext):
        pass

    # Exit a parse tree produced by ASLParser#timestamp_jsonata.
    def exitTimestamp_jsonata(self, ctx: ASLParser.Timestamp_jsonataContext):
        pass

    # Enter a parse tree produced by ASLParser#timestamp_string.
    def enterTimestamp_string(self, ctx: ASLParser.Timestamp_stringContext):
        pass

    # Exit a parse tree produced by ASLParser#timestamp_string.
    def exitTimestamp_string(self, ctx: ASLParser.Timestamp_stringContext):
        pass

    # Enter a parse tree produced by ASLParser#timestamp_path_decl_var.
    def enterTimestamp_path_decl_var(
        self, ctx: ASLParser.Timestamp_path_decl_varContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#timestamp_path_decl_var.
    def exitTimestamp_path_decl_var(
        self, ctx: ASLParser.Timestamp_path_decl_varContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#timestamp_path_decl_value.
    def enterTimestamp_path_decl_value(
        self, ctx: ASLParser.Timestamp_path_decl_valueContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#timestamp_path_decl_value.
    def exitTimestamp_path_decl_value(
        self, ctx: ASLParser.Timestamp_path_decl_valueContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#items_array.
    def enterItems_array(self, ctx: ASLParser.Items_arrayContext):
        pass

    # Exit a parse tree produced by ASLParser#items_array.
    def exitItems_array(self, ctx: ASLParser.Items_arrayContext):
        pass

    # Enter a parse tree produced by ASLParser#items_jsonata.
    def enterItems_jsonata(self, ctx: ASLParser.Items_jsonataContext):
        pass

    # Exit a parse tree produced by ASLParser#items_jsonata.
    def exitItems_jsonata(self, ctx: ASLParser.Items_jsonataContext):
        pass

    # Enter a parse tree produced by ASLParser#items_path_decl_path_context_object.
    def enterItems_path_decl_path_context_object(
        self, ctx: ASLParser.Items_path_decl_path_context_objectContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#items_path_decl_path_context_object.
    def exitItems_path_decl_path_context_object(
        self, ctx: ASLParser.Items_path_decl_path_context_objectContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#items_path_decl_path_var.
    def enterItems_path_decl_path_var(
        self, ctx: ASLParser.Items_path_decl_path_varContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#items_path_decl_path_var.
    def exitItems_path_decl_path_var(
        self, ctx: ASLParser.Items_path_decl_path_varContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#items_path_decl_path.
    def enterItems_path_decl_path(self, ctx: ASLParser.Items_path_decl_pathContext):
        pass

    # Exit a parse tree produced by ASLParser#items_path_decl_path.
    def exitItems_path_decl_path(self, ctx: ASLParser.Items_path_decl_pathContext):
        pass

    # Enter a parse tree produced by ASLParser#max_concurrency_jsonata.
    def enterMax_concurrency_jsonata(
        self, ctx: ASLParser.Max_concurrency_jsonataContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#max_concurrency_jsonata.
    def exitMax_concurrency_jsonata(
        self, ctx: ASLParser.Max_concurrency_jsonataContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#max_concurrency_int.
    def enterMax_concurrency_int(self, ctx: ASLParser.Max_concurrency_intContext):
        pass

    # Exit a parse tree produced by ASLParser#max_concurrency_int.
    def exitMax_concurrency_int(self, ctx: ASLParser.Max_concurrency_intContext):
        pass

    # Enter a parse tree produced by ASLParser#max_concurrency_path_var.
    def enterMax_concurrency_path_var(
        self, ctx: ASLParser.Max_concurrency_path_varContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#max_concurrency_path_var.
    def exitMax_concurrency_path_var(
        self, ctx: ASLParser.Max_concurrency_path_varContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#max_concurrency_path.
    def enterMax_concurrency_path(self, ctx: ASLParser.Max_concurrency_pathContext):
        pass

    # Exit a parse tree produced by ASLParser#max_concurrency_path.
    def exitMax_concurrency_path(self, ctx: ASLParser.Max_concurrency_pathContext):
        pass

    # Enter a parse tree produced by ASLParser#parameters_decl.
    def enterParameters_decl(self, ctx: ASLParser.Parameters_declContext):
        pass

    # Exit a parse tree produced by ASLParser#parameters_decl.
    def exitParameters_decl(self, ctx: ASLParser.Parameters_declContext):
        pass

    # Enter a parse tree produced by ASLParser#credentials_decl.
    def enterCredentials_decl(self, ctx: ASLParser.Credentials_declContext):
        pass

    # Exit a parse tree produced by ASLParser#credentials_decl.
    def exitCredentials_decl(self, ctx: ASLParser.Credentials_declContext):
        pass

    # Enter a parse tree produced by ASLParser#role_arn_jsonata.
    def enterRole_arn_jsonata(self, ctx: ASLParser.Role_arn_jsonataContext):
        pass

    # Exit a parse tree produced by ASLParser#role_arn_jsonata.
    def exitRole_arn_jsonata(self, ctx: ASLParser.Role_arn_jsonataContext):
        pass

    # Enter a parse tree produced by ASLParser#role_arn_path.
    def enterRole_arn_path(self, ctx: ASLParser.Role_arn_pathContext):
        pass

    # Exit a parse tree produced by ASLParser#role_arn_path.
    def exitRole_arn_path(self, ctx: ASLParser.Role_arn_pathContext):
        pass

    # Enter a parse tree produced by ASLParser#role_arn_path_context_obj.
    def enterRole_arn_path_context_obj(
        self, ctx: ASLParser.Role_arn_path_context_objContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#role_arn_path_context_obj.
    def exitRole_arn_path_context_obj(
        self, ctx: ASLParser.Role_arn_path_context_objContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#role_arn_intrinsic_func.
    def enterRole_arn_intrinsic_func(
        self, ctx: ASLParser.Role_arn_intrinsic_funcContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#role_arn_intrinsic_func.
    def exitRole_arn_intrinsic_func(
        self, ctx: ASLParser.Role_arn_intrinsic_funcContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#role_arn_var.
    def enterRole_arn_var(self, ctx: ASLParser.Role_arn_varContext):
        pass

    # Exit a parse tree produced by ASLParser#role_arn_var.
    def exitRole_arn_var(self, ctx: ASLParser.Role_arn_varContext):
        pass

    # Enter a parse tree produced by ASLParser#role_arn_str.
    def enterRole_arn_str(self, ctx: ASLParser.Role_arn_strContext):
        pass

    # Exit a parse tree produced by ASLParser#role_arn_str.
    def exitRole_arn_str(self, ctx: ASLParser.Role_arn_strContext):
        pass

    # Enter a parse tree produced by ASLParser#timeout_seconds_jsonata.
    def enterTimeout_seconds_jsonata(
        self, ctx: ASLParser.Timeout_seconds_jsonataContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#timeout_seconds_jsonata.
    def exitTimeout_seconds_jsonata(
        self, ctx: ASLParser.Timeout_seconds_jsonataContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#timeout_seconds_int.
    def enterTimeout_seconds_int(self, ctx: ASLParser.Timeout_seconds_intContext):
        pass

    # Exit a parse tree produced by ASLParser#timeout_seconds_int.
    def exitTimeout_seconds_int(self, ctx: ASLParser.Timeout_seconds_intContext):
        pass

    # Enter a parse tree produced by ASLParser#timeout_seconds_path_decl_var.
    def enterTimeout_seconds_path_decl_var(
        self, ctx: ASLParser.Timeout_seconds_path_decl_varContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#timeout_seconds_path_decl_var.
    def exitTimeout_seconds_path_decl_var(
        self, ctx: ASLParser.Timeout_seconds_path_decl_varContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#timeout_seconds_path_decl_path.
    def enterTimeout_seconds_path_decl_path(
        self, ctx: ASLParser.Timeout_seconds_path_decl_pathContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#timeout_seconds_path_decl_path.
    def exitTimeout_seconds_path_decl_path(
        self, ctx: ASLParser.Timeout_seconds_path_decl_pathContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#heartbeat_seconds_jsonata.
    def enterHeartbeat_seconds_jsonata(
        self, ctx: ASLParser.Heartbeat_seconds_jsonataContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#heartbeat_seconds_jsonata.
    def exitHeartbeat_seconds_jsonata(
        self, ctx: ASLParser.Heartbeat_seconds_jsonataContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#heartbeat_seconds_int.
    def enterHeartbeat_seconds_int(self, ctx: ASLParser.Heartbeat_seconds_intContext):
        pass

    # Exit a parse tree produced by ASLParser#heartbeat_seconds_int.
    def exitHeartbeat_seconds_int(self, ctx: ASLParser.Heartbeat_seconds_intContext):
        pass

    # Enter a parse tree produced by ASLParser#heartbeat_seconds_path_decl_var.
    def enterHeartbeat_seconds_path_decl_var(
        self, ctx: ASLParser.Heartbeat_seconds_path_decl_varContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#heartbeat_seconds_path_decl_var.
    def exitHeartbeat_seconds_path_decl_var(
        self, ctx: ASLParser.Heartbeat_seconds_path_decl_varContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#heartbeat_seconds_path_decl_path.
    def enterHeartbeat_seconds_path_decl_path(
        self, ctx: ASLParser.Heartbeat_seconds_path_decl_pathContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#heartbeat_seconds_path_decl_path.
    def exitHeartbeat_seconds_path_decl_path(
        self, ctx: ASLParser.Heartbeat_seconds_path_decl_pathContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#variable_sample.
    def enterVariable_sample(self, ctx: ASLParser.Variable_sampleContext):
        pass

    # Exit a parse tree produced by ASLParser#variable_sample.
    def exitVariable_sample(self, ctx: ASLParser.Variable_sampleContext):
        pass

    # Enter a parse tree produced by ASLParser#payload_tmpl_decl.
    def enterPayload_tmpl_decl(self, ctx: ASLParser.Payload_tmpl_declContext):
        pass

    # Exit a parse tree produced by ASLParser#payload_tmpl_decl.
    def exitPayload_tmpl_decl(self, ctx: ASLParser.Payload_tmpl_declContext):
        pass

    # Enter a parse tree produced by ASLParser#payload_binding_path.
    def enterPayload_binding_path(self, ctx: ASLParser.Payload_binding_pathContext):
        pass

    # Exit a parse tree produced by ASLParser#payload_binding_path.
    def exitPayload_binding_path(self, ctx: ASLParser.Payload_binding_pathContext):
        pass

    # Enter a parse tree produced by ASLParser#payload_binding_path_context_obj.
    def enterPayload_binding_path_context_obj(
        self, ctx: ASLParser.Payload_binding_path_context_objContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#payload_binding_path_context_obj.
    def exitPayload_binding_path_context_obj(
        self, ctx: ASLParser.Payload_binding_path_context_objContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#payload_binding_intrinsic_func.
    def enterPayload_binding_intrinsic_func(
        self, ctx: ASLParser.Payload_binding_intrinsic_funcContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#payload_binding_intrinsic_func.
    def exitPayload_binding_intrinsic_func(
        self, ctx: ASLParser.Payload_binding_intrinsic_funcContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#payload_binding_var.
    def enterPayload_binding_var(self, ctx: ASLParser.Payload_binding_varContext):
        pass

    # Exit a parse tree produced by ASLParser#payload_binding_var.
    def exitPayload_binding_var(self, ctx: ASLParser.Payload_binding_varContext):
        pass

    # Enter a parse tree produced by ASLParser#payload_binding_value.
    def enterPayload_binding_value(self, ctx: ASLParser.Payload_binding_valueContext):
        pass

    # Exit a parse tree produced by ASLParser#payload_binding_value.
    def exitPayload_binding_value(self, ctx: ASLParser.Payload_binding_valueContext):
        pass

    # Enter a parse tree produced by ASLParser#payload_arr_decl.
    def enterPayload_arr_decl(self, ctx: ASLParser.Payload_arr_declContext):
        pass

    # Exit a parse tree produced by ASLParser#payload_arr_decl.
    def exitPayload_arr_decl(self, ctx: ASLParser.Payload_arr_declContext):
        pass

    # Enter a parse tree produced by ASLParser#payload_value_decl.
    def enterPayload_value_decl(self, ctx: ASLParser.Payload_value_declContext):
        pass

    # Exit a parse tree produced by ASLParser#payload_value_decl.
    def exitPayload_value_decl(self, ctx: ASLParser.Payload_value_declContext):
        pass

    # Enter a parse tree produced by ASLParser#payload_value_float.
    def enterPayload_value_float(self, ctx: ASLParser.Payload_value_floatContext):
        pass

    # Exit a parse tree produced by ASLParser#payload_value_float.
    def exitPayload_value_float(self, ctx: ASLParser.Payload_value_floatContext):
        pass

    # Enter a parse tree produced by ASLParser#payload_value_int.
    def enterPayload_value_int(self, ctx: ASLParser.Payload_value_intContext):
        pass

    # Exit a parse tree produced by ASLParser#payload_value_int.
    def exitPayload_value_int(self, ctx: ASLParser.Payload_value_intContext):
        pass

    # Enter a parse tree produced by ASLParser#payload_value_bool.
    def enterPayload_value_bool(self, ctx: ASLParser.Payload_value_boolContext):
        pass

    # Exit a parse tree produced by ASLParser#payload_value_bool.
    def exitPayload_value_bool(self, ctx: ASLParser.Payload_value_boolContext):
        pass

    # Enter a parse tree produced by ASLParser#payload_value_null.
    def enterPayload_value_null(self, ctx: ASLParser.Payload_value_nullContext):
        pass

    # Exit a parse tree produced by ASLParser#payload_value_null.
    def exitPayload_value_null(self, ctx: ASLParser.Payload_value_nullContext):
        pass

    # Enter a parse tree produced by ASLParser#payload_value_str.
    def enterPayload_value_str(self, ctx: ASLParser.Payload_value_strContext):
        pass

    # Exit a parse tree produced by ASLParser#payload_value_str.
    def exitPayload_value_str(self, ctx: ASLParser.Payload_value_strContext):
        pass

    # Enter a parse tree produced by ASLParser#assign_decl.
    def enterAssign_decl(self, ctx: ASLParser.Assign_declContext):
        pass

    # Exit a parse tree produced by ASLParser#assign_decl.
    def exitAssign_decl(self, ctx: ASLParser.Assign_declContext):
        pass

    # Enter a parse tree produced by ASLParser#assign_decl_body.
    def enterAssign_decl_body(self, ctx: ASLParser.Assign_decl_bodyContext):
        pass

    # Exit a parse tree produced by ASLParser#assign_decl_body.
    def exitAssign_decl_body(self, ctx: ASLParser.Assign_decl_bodyContext):
        pass

    # Enter a parse tree produced by ASLParser#assign_decl_binding.
    def enterAssign_decl_binding(self, ctx: ASLParser.Assign_decl_bindingContext):
        pass

    # Exit a parse tree produced by ASLParser#assign_decl_binding.
    def exitAssign_decl_binding(self, ctx: ASLParser.Assign_decl_bindingContext):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_value_object.
    def enterAssign_template_value_object(
        self, ctx: ASLParser.Assign_template_value_objectContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_value_object.
    def exitAssign_template_value_object(
        self, ctx: ASLParser.Assign_template_value_objectContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_binding_path.
    def enterAssign_template_binding_path(
        self, ctx: ASLParser.Assign_template_binding_pathContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_binding_path.
    def exitAssign_template_binding_path(
        self, ctx: ASLParser.Assign_template_binding_pathContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_binding_path_context.
    def enterAssign_template_binding_path_context(
        self, ctx: ASLParser.Assign_template_binding_path_contextContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_binding_path_context.
    def exitAssign_template_binding_path_context(
        self, ctx: ASLParser.Assign_template_binding_path_contextContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_binding_var.
    def enterAssign_template_binding_var(
        self, ctx: ASLParser.Assign_template_binding_varContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_binding_var.
    def exitAssign_template_binding_var(
        self, ctx: ASLParser.Assign_template_binding_varContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_binding_intrinsic_func.
    def enterAssign_template_binding_intrinsic_func(
        self, ctx: ASLParser.Assign_template_binding_intrinsic_funcContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_binding_intrinsic_func.
    def exitAssign_template_binding_intrinsic_func(
        self, ctx: ASLParser.Assign_template_binding_intrinsic_funcContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_binding_assign_value.
    def enterAssign_template_binding_assign_value(
        self, ctx: ASLParser.Assign_template_binding_assign_valueContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_binding_assign_value.
    def exitAssign_template_binding_assign_value(
        self, ctx: ASLParser.Assign_template_binding_assign_valueContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_value.
    def enterAssign_template_value(self, ctx: ASLParser.Assign_template_valueContext):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_value.
    def exitAssign_template_value(self, ctx: ASLParser.Assign_template_valueContext):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_value_array.
    def enterAssign_template_value_array(
        self, ctx: ASLParser.Assign_template_value_arrayContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_value_array.
    def exitAssign_template_value_array(
        self, ctx: ASLParser.Assign_template_value_arrayContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_value_terminal_float.
    def enterAssign_template_value_terminal_float(
        self, ctx: ASLParser.Assign_template_value_terminal_floatContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_value_terminal_float.
    def exitAssign_template_value_terminal_float(
        self, ctx: ASLParser.Assign_template_value_terminal_floatContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_value_terminal_int.
    def enterAssign_template_value_terminal_int(
        self, ctx: ASLParser.Assign_template_value_terminal_intContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_value_terminal_int.
    def exitAssign_template_value_terminal_int(
        self, ctx: ASLParser.Assign_template_value_terminal_intContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_value_terminal_bool.
    def enterAssign_template_value_terminal_bool(
        self, ctx: ASLParser.Assign_template_value_terminal_boolContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_value_terminal_bool.
    def exitAssign_template_value_terminal_bool(
        self, ctx: ASLParser.Assign_template_value_terminal_boolContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_value_terminal_null.
    def enterAssign_template_value_terminal_null(
        self, ctx: ASLParser.Assign_template_value_terminal_nullContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_value_terminal_null.
    def exitAssign_template_value_terminal_null(
        self, ctx: ASLParser.Assign_template_value_terminal_nullContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_value_terminal_expression.
    def enterAssign_template_value_terminal_expression(
        self, ctx: ASLParser.Assign_template_value_terminal_expressionContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_value_terminal_expression.
    def exitAssign_template_value_terminal_expression(
        self, ctx: ASLParser.Assign_template_value_terminal_expressionContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#assign_template_value_terminal_str.
    def enterAssign_template_value_terminal_str(
        self, ctx: ASLParser.Assign_template_value_terminal_strContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#assign_template_value_terminal_str.
    def exitAssign_template_value_terminal_str(
        self, ctx: ASLParser.Assign_template_value_terminal_strContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#arguments_object.
    def enterArguments_object(self, ctx: ASLParser.Arguments_objectContext):
        pass

    # Exit a parse tree produced by ASLParser#arguments_object.
    def exitArguments_object(self, ctx: ASLParser.Arguments_objectContext):
        pass

    # Enter a parse tree produced by ASLParser#arguments_expr.
    def enterArguments_expr(self, ctx: ASLParser.Arguments_exprContext):
        pass

    # Exit a parse tree produced by ASLParser#arguments_expr.
    def exitArguments_expr(self, ctx: ASLParser.Arguments_exprContext):
        pass

    # Enter a parse tree produced by ASLParser#output_decl.
    def enterOutput_decl(self, ctx: ASLParser.Output_declContext):
        pass

    # Exit a parse tree produced by ASLParser#output_decl.
    def exitOutput_decl(self, ctx: ASLParser.Output_declContext):
        pass

    # Enter a parse tree produced by ASLParser#jsonata_template_value_object.
    def enterJsonata_template_value_object(
        self, ctx: ASLParser.Jsonata_template_value_objectContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#jsonata_template_value_object.
    def exitJsonata_template_value_object(
        self, ctx: ASLParser.Jsonata_template_value_objectContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#jsonata_template_binding.
    def enterJsonata_template_binding(
        self, ctx: ASLParser.Jsonata_template_bindingContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#jsonata_template_binding.
    def exitJsonata_template_binding(
        self, ctx: ASLParser.Jsonata_template_bindingContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#jsonata_template_value.
    def enterJsonata_template_value(self, ctx: ASLParser.Jsonata_template_valueContext):
        pass

    # Exit a parse tree produced by ASLParser#jsonata_template_value.
    def exitJsonata_template_value(self, ctx: ASLParser.Jsonata_template_valueContext):
        pass

    # Enter a parse tree produced by ASLParser#jsonata_template_value_array.
    def enterJsonata_template_value_array(
        self, ctx: ASLParser.Jsonata_template_value_arrayContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#jsonata_template_value_array.
    def exitJsonata_template_value_array(
        self, ctx: ASLParser.Jsonata_template_value_arrayContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#jsonata_template_value_terminal_float.
    def enterJsonata_template_value_terminal_float(
        self, ctx: ASLParser.Jsonata_template_value_terminal_floatContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#jsonata_template_value_terminal_float.
    def exitJsonata_template_value_terminal_float(
        self, ctx: ASLParser.Jsonata_template_value_terminal_floatContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#jsonata_template_value_terminal_int.
    def enterJsonata_template_value_terminal_int(
        self, ctx: ASLParser.Jsonata_template_value_terminal_intContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#jsonata_template_value_terminal_int.
    def exitJsonata_template_value_terminal_int(
        self, ctx: ASLParser.Jsonata_template_value_terminal_intContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#jsonata_template_value_terminal_bool.
    def enterJsonata_template_value_terminal_bool(
        self, ctx: ASLParser.Jsonata_template_value_terminal_boolContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#jsonata_template_value_terminal_bool.
    def exitJsonata_template_value_terminal_bool(
        self, ctx: ASLParser.Jsonata_template_value_terminal_boolContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#jsonata_template_value_terminal_null.
    def enterJsonata_template_value_terminal_null(
        self, ctx: ASLParser.Jsonata_template_value_terminal_nullContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#jsonata_template_value_terminal_null.
    def exitJsonata_template_value_terminal_null(
        self, ctx: ASLParser.Jsonata_template_value_terminal_nullContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#jsonata_template_value_terminal_expression.
    def enterJsonata_template_value_terminal_expression(
        self, ctx: ASLParser.Jsonata_template_value_terminal_expressionContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#jsonata_template_value_terminal_expression.
    def exitJsonata_template_value_terminal_expression(
        self, ctx: ASLParser.Jsonata_template_value_terminal_expressionContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#jsonata_template_value_terminal_str.
    def enterJsonata_template_value_terminal_str(
        self, ctx: ASLParser.Jsonata_template_value_terminal_strContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#jsonata_template_value_terminal_str.
    def exitJsonata_template_value_terminal_str(
        self, ctx: ASLParser.Jsonata_template_value_terminal_strContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#result_selector_decl.
    def enterResult_selector_decl(self, ctx: ASLParser.Result_selector_declContext):
        pass

    # Exit a parse tree produced by ASLParser#result_selector_decl.
    def exitResult_selector_decl(self, ctx: ASLParser.Result_selector_declContext):
        pass

    # Enter a parse tree produced by ASLParser#state_type.
    def enterState_type(self, ctx: ASLParser.State_typeContext):
        pass

    # Exit a parse tree produced by ASLParser#state_type.
    def exitState_type(self, ctx: ASLParser.State_typeContext):
        pass

    # Enter a parse tree produced by ASLParser#choices_decl.
    def enterChoices_decl(self, ctx: ASLParser.Choices_declContext):
        pass

    # Exit a parse tree produced by ASLParser#choices_decl.
    def exitChoices_decl(self, ctx: ASLParser.Choices_declContext):
        pass

    # Enter a parse tree produced by ASLParser#choice_rule_comparison_variable.
    def enterChoice_rule_comparison_variable(
        self, ctx: ASLParser.Choice_rule_comparison_variableContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#choice_rule_comparison_variable.
    def exitChoice_rule_comparison_variable(
        self, ctx: ASLParser.Choice_rule_comparison_variableContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#choice_rule_comparison_composite.
    def enterChoice_rule_comparison_composite(
        self, ctx: ASLParser.Choice_rule_comparison_compositeContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#choice_rule_comparison_composite.
    def exitChoice_rule_comparison_composite(
        self, ctx: ASLParser.Choice_rule_comparison_compositeContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#comparison_variable_stmt.
    def enterComparison_variable_stmt(
        self, ctx: ASLParser.Comparison_variable_stmtContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#comparison_variable_stmt.
    def exitComparison_variable_stmt(
        self, ctx: ASLParser.Comparison_variable_stmtContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#comparison_composite_stmt.
    def enterComparison_composite_stmt(
        self, ctx: ASLParser.Comparison_composite_stmtContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#comparison_composite_stmt.
    def exitComparison_composite_stmt(
        self, ctx: ASLParser.Comparison_composite_stmtContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#comparison_composite.
    def enterComparison_composite(self, ctx: ASLParser.Comparison_compositeContext):
        pass

    # Exit a parse tree produced by ASLParser#comparison_composite.
    def exitComparison_composite(self, ctx: ASLParser.Comparison_compositeContext):
        pass

    # Enter a parse tree produced by ASLParser#variable_decl_path.
    def enterVariable_decl_path(self, ctx: ASLParser.Variable_decl_pathContext):
        pass

    # Exit a parse tree produced by ASLParser#variable_decl_path.
    def exitVariable_decl_path(self, ctx: ASLParser.Variable_decl_pathContext):
        pass

    # Enter a parse tree produced by ASLParser#variable_decl_var.
    def enterVariable_decl_var(self, ctx: ASLParser.Variable_decl_varContext):
        pass

    # Exit a parse tree produced by ASLParser#variable_decl_var.
    def exitVariable_decl_var(self, ctx: ASLParser.Variable_decl_varContext):
        pass

    # Enter a parse tree produced by ASLParser#variable_decl_path_context_object.
    def enterVariable_decl_path_context_object(
        self, ctx: ASLParser.Variable_decl_path_context_objectContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#variable_decl_path_context_object.
    def exitVariable_decl_path_context_object(
        self, ctx: ASLParser.Variable_decl_path_context_objectContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#condition_lit.
    def enterCondition_lit(self, ctx: ASLParser.Condition_litContext):
        pass

    # Exit a parse tree produced by ASLParser#condition_lit.
    def exitCondition_lit(self, ctx: ASLParser.Condition_litContext):
        pass

    # Enter a parse tree produced by ASLParser#condition_expr.
    def enterCondition_expr(self, ctx: ASLParser.Condition_exprContext):
        pass

    # Exit a parse tree produced by ASLParser#condition_expr.
    def exitCondition_expr(self, ctx: ASLParser.Condition_exprContext):
        pass

    # Enter a parse tree produced by ASLParser#comparison_func_var.
    def enterComparison_func_var(self, ctx: ASLParser.Comparison_func_varContext):
        pass

    # Exit a parse tree produced by ASLParser#comparison_func_var.
    def exitComparison_func_var(self, ctx: ASLParser.Comparison_func_varContext):
        pass

    # Enter a parse tree produced by ASLParser#comparison_func_value.
    def enterComparison_func_value(self, ctx: ASLParser.Comparison_func_valueContext):
        pass

    # Exit a parse tree produced by ASLParser#comparison_func_value.
    def exitComparison_func_value(self, ctx: ASLParser.Comparison_func_valueContext):
        pass

    # Enter a parse tree produced by ASLParser#branches_decl.
    def enterBranches_decl(self, ctx: ASLParser.Branches_declContext):
        pass

    # Exit a parse tree produced by ASLParser#branches_decl.
    def exitBranches_decl(self, ctx: ASLParser.Branches_declContext):
        pass

    # Enter a parse tree produced by ASLParser#item_processor_decl.
    def enterItem_processor_decl(self, ctx: ASLParser.Item_processor_declContext):
        pass

    # Exit a parse tree produced by ASLParser#item_processor_decl.
    def exitItem_processor_decl(self, ctx: ASLParser.Item_processor_declContext):
        pass

    # Enter a parse tree produced by ASLParser#item_processor_item.
    def enterItem_processor_item(self, ctx: ASLParser.Item_processor_itemContext):
        pass

    # Exit a parse tree produced by ASLParser#item_processor_item.
    def exitItem_processor_item(self, ctx: ASLParser.Item_processor_itemContext):
        pass

    # Enter a parse tree produced by ASLParser#processor_config_decl.
    def enterProcessor_config_decl(self, ctx: ASLParser.Processor_config_declContext):
        pass

    # Exit a parse tree produced by ASLParser#processor_config_decl.
    def exitProcessor_config_decl(self, ctx: ASLParser.Processor_config_declContext):
        pass

    # Enter a parse tree produced by ASLParser#processor_config_field.
    def enterProcessor_config_field(self, ctx: ASLParser.Processor_config_fieldContext):
        pass

    # Exit a parse tree produced by ASLParser#processor_config_field.
    def exitProcessor_config_field(self, ctx: ASLParser.Processor_config_fieldContext):
        pass

    # Enter a parse tree produced by ASLParser#mode_decl.
    def enterMode_decl(self, ctx: ASLParser.Mode_declContext):
        pass

    # Exit a parse tree produced by ASLParser#mode_decl.
    def exitMode_decl(self, ctx: ASLParser.Mode_declContext):
        pass

    # Enter a parse tree produced by ASLParser#mode_type.
    def enterMode_type(self, ctx: ASLParser.Mode_typeContext):
        pass

    # Exit a parse tree produced by ASLParser#mode_type.
    def exitMode_type(self, ctx: ASLParser.Mode_typeContext):
        pass

    # Enter a parse tree produced by ASLParser#execution_decl.
    def enterExecution_decl(self, ctx: ASLParser.Execution_declContext):
        pass

    # Exit a parse tree produced by ASLParser#execution_decl.
    def exitExecution_decl(self, ctx: ASLParser.Execution_declContext):
        pass

    # Enter a parse tree produced by ASLParser#execution_type.
    def enterExecution_type(self, ctx: ASLParser.Execution_typeContext):
        pass

    # Exit a parse tree produced by ASLParser#execution_type.
    def exitExecution_type(self, ctx: ASLParser.Execution_typeContext):
        pass

    # Enter a parse tree produced by ASLParser#iterator_decl.
    def enterIterator_decl(self, ctx: ASLParser.Iterator_declContext):
        pass

    # Exit a parse tree produced by ASLParser#iterator_decl.
    def exitIterator_decl(self, ctx: ASLParser.Iterator_declContext):
        pass

    # Enter a parse tree produced by ASLParser#iterator_decl_item.
    def enterIterator_decl_item(self, ctx: ASLParser.Iterator_decl_itemContext):
        pass

    # Exit a parse tree produced by ASLParser#iterator_decl_item.
    def exitIterator_decl_item(self, ctx: ASLParser.Iterator_decl_itemContext):
        pass

    # Enter a parse tree produced by ASLParser#item_selector_decl.
    def enterItem_selector_decl(self, ctx: ASLParser.Item_selector_declContext):
        pass

    # Exit a parse tree produced by ASLParser#item_selector_decl.
    def exitItem_selector_decl(self, ctx: ASLParser.Item_selector_declContext):
        pass

    # Enter a parse tree produced by ASLParser#item_reader_decl.
    def enterItem_reader_decl(self, ctx: ASLParser.Item_reader_declContext):
        pass

    # Exit a parse tree produced by ASLParser#item_reader_decl.
    def exitItem_reader_decl(self, ctx: ASLParser.Item_reader_declContext):
        pass

    # Enter a parse tree produced by ASLParser#items_reader_field.
    def enterItems_reader_field(self, ctx: ASLParser.Items_reader_fieldContext):
        pass

    # Exit a parse tree produced by ASLParser#items_reader_field.
    def exitItems_reader_field(self, ctx: ASLParser.Items_reader_fieldContext):
        pass

    # Enter a parse tree produced by ASLParser#reader_config_decl.
    def enterReader_config_decl(self, ctx: ASLParser.Reader_config_declContext):
        pass

    # Exit a parse tree produced by ASLParser#reader_config_decl.
    def exitReader_config_decl(self, ctx: ASLParser.Reader_config_declContext):
        pass

    # Enter a parse tree produced by ASLParser#reader_config_field.
    def enterReader_config_field(self, ctx: ASLParser.Reader_config_fieldContext):
        pass

    # Exit a parse tree produced by ASLParser#reader_config_field.
    def exitReader_config_field(self, ctx: ASLParser.Reader_config_fieldContext):
        pass

    # Enter a parse tree produced by ASLParser#input_type_decl.
    def enterInput_type_decl(self, ctx: ASLParser.Input_type_declContext):
        pass

    # Exit a parse tree produced by ASLParser#input_type_decl.
    def exitInput_type_decl(self, ctx: ASLParser.Input_type_declContext):
        pass

    # Enter a parse tree produced by ASLParser#csv_header_location_decl.
    def enterCsv_header_location_decl(
        self, ctx: ASLParser.Csv_header_location_declContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#csv_header_location_decl.
    def exitCsv_header_location_decl(
        self, ctx: ASLParser.Csv_header_location_declContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#csv_headers_decl.
    def enterCsv_headers_decl(self, ctx: ASLParser.Csv_headers_declContext):
        pass

    # Exit a parse tree produced by ASLParser#csv_headers_decl.
    def exitCsv_headers_decl(self, ctx: ASLParser.Csv_headers_declContext):
        pass

    # Enter a parse tree produced by ASLParser#max_items_jsonata.
    def enterMax_items_jsonata(self, ctx: ASLParser.Max_items_jsonataContext):
        pass

    # Exit a parse tree produced by ASLParser#max_items_jsonata.
    def exitMax_items_jsonata(self, ctx: ASLParser.Max_items_jsonataContext):
        pass

    # Enter a parse tree produced by ASLParser#max_items_int.
    def enterMax_items_int(self, ctx: ASLParser.Max_items_intContext):
        pass

    # Exit a parse tree produced by ASLParser#max_items_int.
    def exitMax_items_int(self, ctx: ASLParser.Max_items_intContext):
        pass

    # Enter a parse tree produced by ASLParser#max_items_path_var.
    def enterMax_items_path_var(self, ctx: ASLParser.Max_items_path_varContext):
        pass

    # Exit a parse tree produced by ASLParser#max_items_path_var.
    def exitMax_items_path_var(self, ctx: ASLParser.Max_items_path_varContext):
        pass

    # Enter a parse tree produced by ASLParser#max_items_path.
    def enterMax_items_path(self, ctx: ASLParser.Max_items_pathContext):
        pass

    # Exit a parse tree produced by ASLParser#max_items_path.
    def exitMax_items_path(self, ctx: ASLParser.Max_items_pathContext):
        pass

    # Enter a parse tree produced by ASLParser#tolerated_failure_count_jsonata.
    def enterTolerated_failure_count_jsonata(
        self, ctx: ASLParser.Tolerated_failure_count_jsonataContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#tolerated_failure_count_jsonata.
    def exitTolerated_failure_count_jsonata(
        self, ctx: ASLParser.Tolerated_failure_count_jsonataContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#tolerated_failure_count_int.
    def enterTolerated_failure_count_int(
        self, ctx: ASLParser.Tolerated_failure_count_intContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#tolerated_failure_count_int.
    def exitTolerated_failure_count_int(
        self, ctx: ASLParser.Tolerated_failure_count_intContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#tolerated_failure_count_path_var.
    def enterTolerated_failure_count_path_var(
        self, ctx: ASLParser.Tolerated_failure_count_path_varContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#tolerated_failure_count_path_var.
    def exitTolerated_failure_count_path_var(
        self, ctx: ASLParser.Tolerated_failure_count_path_varContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#tolerated_failure_count_path.
    def enterTolerated_failure_count_path(
        self, ctx: ASLParser.Tolerated_failure_count_pathContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#tolerated_failure_count_path.
    def exitTolerated_failure_count_path(
        self, ctx: ASLParser.Tolerated_failure_count_pathContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#tolerated_failure_percentage_jsonata.
    def enterTolerated_failure_percentage_jsonata(
        self, ctx: ASLParser.Tolerated_failure_percentage_jsonataContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#tolerated_failure_percentage_jsonata.
    def exitTolerated_failure_percentage_jsonata(
        self, ctx: ASLParser.Tolerated_failure_percentage_jsonataContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#tolerated_failure_percentage_number.
    def enterTolerated_failure_percentage_number(
        self, ctx: ASLParser.Tolerated_failure_percentage_numberContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#tolerated_failure_percentage_number.
    def exitTolerated_failure_percentage_number(
        self, ctx: ASLParser.Tolerated_failure_percentage_numberContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#tolerated_failure_percentage_path_var.
    def enterTolerated_failure_percentage_path_var(
        self, ctx: ASLParser.Tolerated_failure_percentage_path_varContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#tolerated_failure_percentage_path_var.
    def exitTolerated_failure_percentage_path_var(
        self, ctx: ASLParser.Tolerated_failure_percentage_path_varContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#tolerated_failure_percentage_path.
    def enterTolerated_failure_percentage_path(
        self, ctx: ASLParser.Tolerated_failure_percentage_pathContext
    ):
        pass

    # Exit a parse tree produced by ASLParser#tolerated_failure_percentage_path.
    def exitTolerated_failure_percentage_path(
        self, ctx: ASLParser.Tolerated_failure_percentage_pathContext
    ):
        pass

    # Enter a parse tree produced by ASLParser#label_decl.
    def enterLabel_decl(self, ctx: ASLParser.Label_declContext):
        pass

    # Exit a parse tree produced by ASLParser#label_decl.
    def exitLabel_decl(self, ctx: ASLParser.Label_declContext):
        pass

    # Enter a parse tree produced by ASLParser#result_writer_decl.
    def enterResult_writer_decl(self, ctx: ASLParser.Result_writer_declContext):
        pass

    # Exit a parse tree produced by ASLParser#result_writer_decl.
    def exitResult_writer_decl(self, ctx: ASLParser.Result_writer_declContext):
        pass

    # Enter a parse tree produced by ASLParser#result_writer_field.
    def enterResult_writer_field(self, ctx: ASLParser.Result_writer_fieldContext):
        pass

    # Exit a parse tree produced by ASLParser#result_writer_field.
    def exitResult_writer_field(self, ctx: ASLParser.Result_writer_fieldContext):
        pass

    # Enter a parse tree produced by ASLParser#retry_decl.
    def enterRetry_decl(self, ctx: ASLParser.Retry_declContext):
        pass

    # Exit a parse tree produced by ASLParser#retry_decl.
    def exitRetry_decl(self, ctx: ASLParser.Retry_declContext):
        pass

    # Enter a parse tree produced by ASLParser#retrier_decl.
    def enterRetrier_decl(self, ctx: ASLParser.Retrier_declContext):
        pass

    # Exit a parse tree produced by ASLParser#retrier_decl.
    def exitRetrier_decl(self, ctx: ASLParser.Retrier_declContext):
        pass

    # Enter a parse tree produced by ASLParser#retrier_stmt.
    def enterRetrier_stmt(self, ctx: ASLParser.Retrier_stmtContext):
        pass

    # Exit a parse tree produced by ASLParser#retrier_stmt.
    def exitRetrier_stmt(self, ctx: ASLParser.Retrier_stmtContext):
        pass

    # Enter a parse tree produced by ASLParser#error_equals_decl.
    def enterError_equals_decl(self, ctx: ASLParser.Error_equals_declContext):
        pass

    # Exit a parse tree produced by ASLParser#error_equals_decl.
    def exitError_equals_decl(self, ctx: ASLParser.Error_equals_declContext):
        pass

    # Enter a parse tree produced by ASLParser#interval_seconds_decl.
    def enterInterval_seconds_decl(self, ctx: ASLParser.Interval_seconds_declContext):
        pass

    # Exit a parse tree produced by ASLParser#interval_seconds_decl.
    def exitInterval_seconds_decl(self, ctx: ASLParser.Interval_seconds_declContext):
        pass

    # Enter a parse tree produced by ASLParser#max_attempts_decl.
    def enterMax_attempts_decl(self, ctx: ASLParser.Max_attempts_declContext):
        pass

    # Exit a parse tree produced by ASLParser#max_attempts_decl.
    def exitMax_attempts_decl(self, ctx: ASLParser.Max_attempts_declContext):
        pass

    # Enter a parse tree produced by ASLParser#backoff_rate_decl.
    def enterBackoff_rate_decl(self, ctx: ASLParser.Backoff_rate_declContext):
        pass

    # Exit a parse tree produced by ASLParser#backoff_rate_decl.
    def exitBackoff_rate_decl(self, ctx: ASLParser.Backoff_rate_declContext):
        pass

    # Enter a parse tree produced by ASLParser#max_delay_seconds_decl.
    def enterMax_delay_seconds_decl(self, ctx: ASLParser.Max_delay_seconds_declContext):
        pass

    # Exit a parse tree produced by ASLParser#max_delay_seconds_decl.
    def exitMax_delay_seconds_decl(self, ctx: ASLParser.Max_delay_seconds_declContext):
        pass

    # Enter a parse tree produced by ASLParser#jitter_strategy_decl.
    def enterJitter_strategy_decl(self, ctx: ASLParser.Jitter_strategy_declContext):
        pass

    # Exit a parse tree produced by ASLParser#jitter_strategy_decl.
    def exitJitter_strategy_decl(self, ctx: ASLParser.Jitter_strategy_declContext):
        pass

    # Enter a parse tree produced by ASLParser#catch_decl.
    def enterCatch_decl(self, ctx: ASLParser.Catch_declContext):
        pass

    # Exit a parse tree produced by ASLParser#catch_decl.
    def exitCatch_decl(self, ctx: ASLParser.Catch_declContext):
        pass

    # Enter a parse tree produced by ASLParser#catcher_decl.
    def enterCatcher_decl(self, ctx: ASLParser.Catcher_declContext):
        pass

    # Exit a parse tree produced by ASLParser#catcher_decl.
    def exitCatcher_decl(self, ctx: ASLParser.Catcher_declContext):
        pass

    # Enter a parse tree produced by ASLParser#catcher_stmt.
    def enterCatcher_stmt(self, ctx: ASLParser.Catcher_stmtContext):
        pass

    # Exit a parse tree produced by ASLParser#catcher_stmt.
    def exitCatcher_stmt(self, ctx: ASLParser.Catcher_stmtContext):
        pass

    # Enter a parse tree produced by ASLParser#comparison_op.
    def enterComparison_op(self, ctx: ASLParser.Comparison_opContext):
        pass

    # Exit a parse tree produced by ASLParser#comparison_op.
    def exitComparison_op(self, ctx: ASLParser.Comparison_opContext):
        pass

    # Enter a parse tree produced by ASLParser#choice_operator.
    def enterChoice_operator(self, ctx: ASLParser.Choice_operatorContext):
        pass

    # Exit a parse tree produced by ASLParser#choice_operator.
    def exitChoice_operator(self, ctx: ASLParser.Choice_operatorContext):
        pass

    # Enter a parse tree produced by ASLParser#states_error_name.
    def enterStates_error_name(self, ctx: ASLParser.States_error_nameContext):
        pass

    # Exit a parse tree produced by ASLParser#states_error_name.
    def exitStates_error_name(self, ctx: ASLParser.States_error_nameContext):
        pass

    # Enter a parse tree produced by ASLParser#error_name.
    def enterError_name(self, ctx: ASLParser.Error_nameContext):
        pass

    # Exit a parse tree produced by ASLParser#error_name.
    def exitError_name(self, ctx: ASLParser.Error_nameContext):
        pass

    # Enter a parse tree produced by ASLParser#json_obj_decl.
    def enterJson_obj_decl(self, ctx: ASLParser.Json_obj_declContext):
        pass

    # Exit a parse tree produced by ASLParser#json_obj_decl.
    def exitJson_obj_decl(self, ctx: ASLParser.Json_obj_declContext):
        pass

    # Enter a parse tree produced by ASLParser#json_binding.
    def enterJson_binding(self, ctx: ASLParser.Json_bindingContext):
        pass

    # Exit a parse tree produced by ASLParser#json_binding.
    def exitJson_binding(self, ctx: ASLParser.Json_bindingContext):
        pass

    # Enter a parse tree produced by ASLParser#json_arr_decl.
    def enterJson_arr_decl(self, ctx: ASLParser.Json_arr_declContext):
        pass

    # Exit a parse tree produced by ASLParser#json_arr_decl.
    def exitJson_arr_decl(self, ctx: ASLParser.Json_arr_declContext):
        pass

    # Enter a parse tree produced by ASLParser#json_value_decl.
    def enterJson_value_decl(self, ctx: ASLParser.Json_value_declContext):
        pass

    # Exit a parse tree produced by ASLParser#json_value_decl.
    def exitJson_value_decl(self, ctx: ASLParser.Json_value_declContext):
        pass

    # Enter a parse tree produced by ASLParser#keyword_or_string.
    def enterKeyword_or_string(self, ctx: ASLParser.Keyword_or_stringContext):
        pass

    # Exit a parse tree produced by ASLParser#keyword_or_string.
    def exitKeyword_or_string(self, ctx: ASLParser.Keyword_or_stringContext):
        pass


del ASLParser



import datetime
from threading import Thread
from typing import Dict, Optional

from moto.stepfunctions.parser.api import (
    Arn,
    ExecutionStartedEventDetails,
    HistoryEventExecutionDataDetails,
    HistoryEventType,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.evaluation_details import EvaluationDetails
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.eval.event.event_manager import (
    EventHistoryContext,
)
from moto.stepfunctions.parser.asl.eval.event.logging import (
    CloudWatchLoggingSession,
)
from moto.stepfunctions.parser.asl.eval.states import (
    ContextObjectData,
    ExecutionData,
    StateMachineData,
)
from moto.stepfunctions.parser.asl.parse.asl_parser import AmazonStateLanguageParser
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str
from moto.stepfunctions.parser.backend.activity import Activity
from moto.stepfunctions.parser.backend.execution_worker_comm import (
    ExecutionWorkerCommunication,
)
from moto.stepfunctions.parser.utils import TMP_THREADS


class ExecutionWorker:
    _evaluation_details: EvaluationDetails
    _execution_communication: ExecutionWorkerCommunication
    _cloud_watch_logging_session: Optional[CloudWatchLoggingSession]
    _activity_store: Dict[Arn, Activity]

    env: Optional[Environment]

    def __init__(
        self,
        evaluation_details: EvaluationDetails,
        exec_comm: ExecutionWorkerCommunication,
        cloud_watch_logging_session: Optional[CloudWatchLoggingSession],
        activity_store: Dict[Arn, Activity],
    ):
        self._evaluation_details = evaluation_details
        self._execution_communication = exec_comm
        self._cloud_watch_logging_session = cloud_watch_logging_session
        self._activity_store = activity_store
        self.env = None

    def _get_evaluation_entrypoint(self) -> EvalComponent:
        return AmazonStateLanguageParser.parse(
            self._evaluation_details.state_machine_details.definition
        )[0]

    def _get_evaluation_environment(self) -> Environment:
        return Environment(
            aws_execution_details=self._evaluation_details.aws_execution_details,
            execution_type=self._evaluation_details.state_machine_details.typ,
            context=ContextObjectData(
                Execution=ExecutionData(
                    Id=self._evaluation_details.execution_details.arn,
                    Input=self._evaluation_details.execution_details.inpt,
                    Name=self._evaluation_details.execution_details.name,
                    RoleArn=self._evaluation_details.execution_details.role_arn,
                    StartTime=self._evaluation_details.execution_details.start_time,
                ),
                StateMachine=StateMachineData(
                    Id=self._evaluation_details.state_machine_details.arn,
                    Name=self._evaluation_details.state_machine_details.name,
                ),
            ),
            event_history_context=EventHistoryContext.of_program_start(),
            cloud_watch_logging_session=self._cloud_watch_logging_session,
            activity_store=self._activity_store,
        )

    def _execution_logic(self):
        program = self._get_evaluation_entrypoint()
        self.env = self._get_evaluation_environment()

        self.env.event_manager.add_event(
            context=self.env.event_history_context,
            event_type=HistoryEventType.ExecutionStarted,
            event_details=EventDetails(
                executionStartedEventDetails=ExecutionStartedEventDetails(
                    input=to_json_str(self._evaluation_details.execution_details.inpt),
                    inputDetails=HistoryEventExecutionDataDetails(
                        truncated=False
                    ),  # Always False for api calls.
                    roleArn=self._evaluation_details.aws_execution_details.role_arn,
                )
            ),
            update_source_event_id=False,
        )

        program.eval(self.env)

        self._execution_communication.terminated()

    def start(self):
        execution_logic_thread = Thread(target=self._execution_logic, daemon=True)
        TMP_THREADS.append(execution_logic_thread)
        execution_logic_thread.start()

    def stop(
        self, stop_date: datetime.datetime, error: Optional[str], cause: Optional[str]
    ):
        self.env.set_stop(stop_date=stop_date, cause=cause, error=error)


class SyncExecutionWorker(ExecutionWorker):
    def start(self):
        # bypass the native async execution of ASL programs.
        self._execution_logic()
